#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
| `@purpose`: Generate regions of interest that can be used for data processing and analysis.  
| `@date`: Created on Sat May 1 15:12:38 2019  
| `@author`: Semeon Risom  
| `@email`: semeon.risom@gmail.com  
| `@url`: https://semeon.io/d/mdl
"""
# allowed imports
__all__ = ['ROI']

# required external libraries
__required__ = ['opencv-python','psd-tools','matplotlib','Pillow']

# local
from .. import settings

# required for init
from pdb import set_trace as breakpoint
import os
from pathlib import Path
import pandas as pd
import numpy as np

# plot
from PIL import Image
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.patches as patches

# check if psd_tools and cv2 is available
# note: anaconda doesn't have a version of psd_tools or opencv-python (cv2) available (5/1/19), so a workaround 
# when building is to directly install it.
try:
	import psd_tools
	import cv2
except ImportError as e:
	import importlib, sys
	pkg = e.name
	settings.console("No module named '%s'. Installing from PyPI."%(pkg),'red')
	# install
	settings.library([pkg])
	# if not all packages are available, reload module
	for x in [['cv2','opencv-python'],['psd_tools','psd-tools']]:
		# check if module is not available
		if importlib.util.find_spec(x[0]) is None:
			settings.console('%s not available'%(x),'red')
			importlib.reload(sys.modules[__name__])
		# import modules
		else:
			try:
				settings.console(('import %s'%(mod[0]),'blue'))
				globals()[mod[0]] = importlib.import_module(mod[0])
			except ImportError as e:
				settings.console("import %s unsuccessful. Trying to install."%(mod[0]),'red')
				importlib.reload(sys.modules[__name__])

class ROI():
	"""Generate regions of interest that can be used for data processing and analysis."""
	def __init__(self, isMultiprocessing=False, image_path=None, output_path=None, metadata_source=None, 
			  roi_format='both', shape='box', roicolumn='roi', uuid=None, **kwargs):
		"""
        Initiate the mdl.eyetracking.ROI module.

		Parameters
		----------
		isMultiprocessing : :obj:`bool`
			Should the rois be generated using multiprocessing. Default `False`.
		detection : :obj:`str`
			How should the regions of interest be detected. Either manually, through the use of highlight layers, or automatically
			using haar cascades opencv. Default `manual`.
		image_path : :obj:`str`
			Image directory path.
		output_path : :class:`str`
			Path to save data.
		roi_format : :obj:`str` {`raw`,`dataviewer`, `both`}
			Format to export ROIs. Either to 'csv' (`raw`) or to Eyelink DataViewer 'ias' (`dataviewer`) or both (`both`). 
			Default is `raw`. Note: If :code:`roi_format` = `dataviewer`, :code:`shape` must be either be `circle`, `rotate`, or `straight`.
		metadata_source : :class:`str` or :obj:`None`
			Metadata source. If metadata is being read from a spreadsheet, :code:`metadata_source` should be equal to path the to
			the metadata file, else if metadata is embed within the image as a layer name, :code:`metadata_source` = `embedded`.
			Default is `embedded`. For example:
				>>> # if metadata is in PSD images
				>>> metadata = 'embedded'
				>>> # if metadata is an external xlsx file.
				>>> metadata = 'roi/metadata.xlsx'
			Although Photoshop PSD don't directly provide support for metadata. However if each region of interest is stored
			as a seperate layer within a PSD, the layer name can be used to store metadata. To do this, the layer name has
			to be written as delimited text. Our code can read this data and extract relevant metadata. The delimiter can
			be either `;` `,` `|` `\\t` or `\\s` (Delimiter type must be identified when running this code using the
			code:`delimiter` parameter. The default is `;`.). Here's an example using `;` as a delimiter:
				>>> imagename=BM001;roiname=1;feature=lefteye
			Note: whitespace should be avoided from from each layer name. Whitespaces may cause errors during parsing.
		shape : :obj:`str` {`polygon`, `hull`, `circle`, `rotate`, `straight`}
			Shape of machine readable boundaries for region of interest. Default is `straight`. `polygon` creates a Contour
			Approximation and will most closely match the orginal shape of the roi. `hull` creates a Convex Hull, which
			is similar to but not as complex as a Contour Approximation and will include bulges for areas that are convex.
			`circle` creates a mininum enclosing circle. Finally, both `rotate` and `straight` create a Bounding Rectangle,
			with the only difference being compensation for the mininum enclosing area for the box when using `rotate`.
		roicolumn : :obj:`str`
			The name of the label for the region of interest in your metadata. For example you may want to extract the column
			'feature' from your metadata and use this as the label. Default is `roi`.
		uuid : :obj:`list` or :obj:`None`
			Create a unique id by combining a list of existing variables in the metadata. This is recommended
			if :code:`roi_format` == `dataviewer` because of the limited variables allowed for ias files. Default :obj:`None`.
		**kwargs : :obj:`str` or :obj:`None`, optional
			Additional properties. Here's a list of available properties:

			.. list-table::
				:class: kwargs
				:widths: 25 50
				:header-rows: 1

				* - Property
				  - Description
				* - **cores** : :class:`bool`
				  - (if :code:`isMultiprocessing` == `True`) Number of cores to use. Default is total available cores - 1.
				* - **isLibrary** : :class:`bool`
				  - Check if required packages have been installed. Default is :obj:`False`.
				* - **isDebug** : :class:`bool`
				  - Allow flags to be visible. Default is :obj:`False`.
				* - **isDemo** : :class:`bool`
				  - Tests code with in-house images and metadata. Default is :obj:`False`.
				* - **save_data** : :class:`bool`
				  - Save coordinates. Default is :obj:`True.`
				* - **newcolumn** : :class:`dict` {:obj:`str`, :obj:`str`} or :obj:`False`
				  - Add additional column to metadata. This must be in the form of a dict in this form {key: value}. 
					Default is :obj:`False.`
				* - **save_raw_image** : :class:`bool`
				  - Save images. Default is True.
				* - **save_contour_image** : :class:`bool`
				  - Save generated contours as images. Default is :obj:`True`.
				* - **delimiter** : :class:`str`
				  - (if :code:`source` == `psd`) How is metadata delimited, options are: `;` `,` `|` `tab` or `space` Default is `;`.
				* - **screensize** : :class:`list` [:obj:`int`]
				  - Monitor size is being presented. Default is `[1920, 1080]`.
				* - **scale** : :class:`int`
				  - If image is scaled during presentation, set scale. Default is 1.
				* - **center** : :class:`list` [:obj:`int`]
				  - Center point of image, relative to screensize. Default is `[960, 540]`.
				* - **dpi** : :class:`int` or :obj:`None`
				  - (if :code:`save_image` == `True`) Quality of exported images, refers to 'dots per inch'. Default is `300`.

		Attributes
		----------
		shape_d : :class:`str` {`ELLIPSE`, `FREEHAND`, `RECTANGLE`}
			DataViewer ROI shape.
		psd :  `psd_tools.PSDImage <https://psd-tools.readthedocs.io/en/latest/reference/psd_tools.html#psd_tools.PSDImage>`_
			Photoshop PSD/PSB file object. The file should include one layer for each region of interest.
		retval, threshold : :obj:`numpy.ndarray`
			Returns from :ref:`cv2.threshold`. The function applies a fixed-level thresholding to a multiple-channel array.
			`retval` provides an optimal threshold only if :ref:`cv2.THRESH_OTSU` is passed. `threshold` is an image after applying
			a binary threshold (:ref:`cv2.THRESH_BINARY`) removing all greyscale pixels < 127. The output matches the same image
			channel as the original image.
			See `opencv <https://docs.opencv.org/4.0.1/d7/d1b/group__imgproc__misc.html#gae8a4a146d1ca78c626a53577199e9c57>`_ and
			`leanopencv <https://www.learnopencv.com/opencv-threshold-python-cpp>`_ for more information.
		contours, hierarchy : :obj:`numpy.ndarray`
			Returns from :ref:`cv2.findContours`. This function returns contours from the provided binary image (threshold).
			This is used here for later shape detection. `contours` are the detected contours, while hierarchy containing
			information about the image topology.
			See `opencv <https://docs.opencv.org/4.0.1/d3/dc0/group__imgproc__shape.html#gadf1ad6a0b82947fa1fe3c3d497f260e07>`_
			for more information.
		image_contours : :obj:`numpy.ndarray`
			Returns from :ref:`cv2.drawContours`. This draws filled contours from the image.
		image_contours : :obj:`numpy.ndarray`
			Returns from :ref:`cv2.drawContours`. This draws filled contours from the image.

		Raises
		------
		Exception
			[description]
		Exception
			[description]

		Examples
		--------
		>>> from mdl.roi import ROI
		>>> s="/dist/example/raw/"; d="/dist/example/"
		>>> ROI(source=s, output_path=d, shape='box')

		Notes
		-----
		Resources
			- See https://docs.opencv.org/3.1.0/dd/d49/tutorial_py_contour_features.html for more information about each shape.
			- See https://docs.opencv.org/2.4/modules/core/doc/drawing_functions.html for more information about how images are drawn.
			- See https://docs.opencv.org/2.4/modules/imgproc/doc/structural_analysis_and_shape_descriptors.html to understand how bounds are created.
		"""
		# get console and time
		self.console = settings.console
		self.now = settings.time

		# check debug
		self.isDebug = kwargs['isDebug'] if 'isDebug' in kwargs else False

		# check library
		self.isLibrary = kwargs['isLibrary'] if 'isLibrary' in kwargs else False
		if self.isLibrary:
			settings.library(__required__)

		#----parameters
		# demo
		self.isDemo = kwargs['isDemo'] if 'isDemo' in kwargs else False
		# multiprocessing
		self.isMultiprocessing = isMultiprocessing
		self.cores = kwargs['cores'] if 'cores' in kwargs else 'max'
		# if multiprocessing, set number of thread
		# if isMultiprocessing:
		# 	cv2.setNumThreads(0)
		# how to read data
		self.metadata_source = kwargs['metadata_source'] if 'metadata_source' in kwargs else 'embed'
		# how to format rois
		self.roi_format = roi_format
		# delimiter
		self.delimiter = kwargs['delimiter'] if 'delimiter' in kwargs else ';'
		# screensize
		self.screensize = kwargs['screensize'] if 'screensize' in kwargs else [1920, 1080]
		# scale
		self.scale = kwargs['scale'] if 'scale' in kwargs else 1
		# coordinates
		cx = self.screensize[0]/2
		cy = self.screensize[1]/2
		self.coordinates = kwargs['coordinates'] if 'coordinates' in kwargs else [cx, cy]
		# shape
		self.shape = shape if shape in ['polygon', 'hull', 'circle', 'rotate', 'straight'] else 'straight'
		self.shape_d = None #dataviewer shape
		# uuid
		self.uuid = uuid
		# label
		self.roicolumn = roicolumn
		# add column
		self.newcolumn = kwargs['newcolumn'] if 'newcolumn' in kwargs else None
		# dpi
		self.dpi =  kwargs['dpi'] if 'dpi' in kwargs else 300
		# save
		self.save = {}
		# save csv
		self.save['data'] = kwargs['save_data'] if 'save_data' in kwargs else True
		# save contour images
		self.save['contours'] = kwargs['save_contour_image'] if 'save_contour_image' in kwargs else True
		# save raw images
		self.save['raw'] = kwargs['save_contour_image'] if 'save_contour_image' in kwargs else True

		#----shape
		# check if trying to do complex ROI using dataviewer
		if (self.shape in ['polygon', 'hull']) and (self.roi_format == "dataviewer"):
			raise Exception ("Cannot use shape %s when exporting for DataViewer. \
					Please use either 'circle', 'rotate', or 'straight' instead, or set roi_format == 'raw'."%(shape))

		#----directory
		if self.isDemo is True:
			import sys
			path = Path(sys.argv[0]).parent
			self.image_path = "%s/raw/"%(path)
			self.output_path = "%s/output/"%(path)
			metadata_source = "%s/metadata.xlsx"%(path)
		else:
			self.image_path = image_path
			self.output_path = output_path
		# set directory of files
		self.directory = [x for x in Path(self.image_path).glob("*.psd") if x.is_file()]

		#----read metadata file (if metadata is not None)
		if metadata_source is not "embedded":
			self.metadata_source = metadata_source
			_type = Path(self.metadata_source).suffix
			if _type == ".csv": self.metadata_all = pd.read_csv(self.metadata_source)
			elif _type == ".xlsx": self.metadata_all = pd.read_excel(self.metadata_source)

			# check if metadata is empty
			if self.metadata_all.empty:
				raise Exception('No data for file: %s'%(self.metadata_source))

	def process_metadata(self, imagename, layer):
		"""[summary]

		Parameters
		----------
		imagename : [type]
			[description]
		layer : [type]
			[description]

		Returns
		-------
		[type]
			[description]
		[type]
			[description]
		[type]
			[description]
		"""

		#----prepare metadata
		# if metadata is stored in image files directly
		if self.metadata_source == 'embedded':
			metadata = pd.DataFrame(data=(item.split("=") for item in layer.name.split(self.delimiter)),columns=['key','value'])
			metadata.set_index('key', inplace=True)
			metadata.loc['name']['value'] = metadata.loc['name']['value'].replace("roi","")
			roiname = metadata.loc['name']['value']
			roilabel = metadata.loc[self.roicolumn]['value']

		# else read metadata from file
		else:
			# get metadata
			roiname = layer.name.strip(' \t\n\r') # strip whitespace
			metadata = self.metadata_all.loc[(self.metadata_all['image'] == imagename) & (self.metadata_all['roi'] == roiname)]
			# if datafame empty
			if metadata.empty:
				message = 'No data for %s:%s (image:roi).'%(imagename, roiname)
				raise Exception(message)
			else:
				roilabel = metadata[self.roicolumn].item()

		# print results
		if self.isDebug:
			self.console('## roiname: %s'%(roiname),'blue')
			self.console('## roilabel: %s'%(roilabel),'green')

		return metadata, roiname, roilabel

	def create_contours(self, image, imagename, roiname):
		"""[summary]

		Parameters
		----------
		image : [type]
			[description]
		imagename : [type]
			[description]
		roiname : [type]
			[description]

		Returns
		-------
		[type]
			[description]

		Raises
		------
		Exception
			[description]
		Exception
			[description]
		"""
		# convert pil image to grayscale (using pil)
		#self.console('test4.0.5', 'red')
		image = image.convert(mode='L')

		# convert to np.array
		#self.console('test4.1.0', 'red')
		image = np.array(image)

		# or convert pil image to grayscale (using cv2)
		#self.console('test4.1.5', 'red')
		#image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

		# threshold the image
		## note: if any pixels that have value higher than 127, assign it to 255. convert to bw for countour and store original
		#self.console('test4.2', 'red')
		retval, threshold = cv2.threshold(image, 127, 255, cv2.THRESH_BINARY)

		# find contour in image
		#self.console('test4.3', 'red')
		## note: if you only want to retrieve the most external contour # use cv.RETR_EXTERNAL
		contours, hierarchy = cv2.findContours(threshold, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

		# if contours empty raise Exception
		#self.console('test4.4', 'red')
		if not bool(contours):
			_err = [imagename, roiname, 'Not able to identify contours']
			raise Exception('%s; %s; %s'%(_err[0],_err[1],_err[2]))

		#------------------------------------------------------------------------for each layer: save images and contours
		#when saving the contours below, only one drawContours function from above can be run
		#any other drawContours function will overlay on the others if multiple functions are run
		#----param
		color = (0,0,255)
		#self.console('test4.5', 'red')

		#----straight bounding box
		if self.shape == 'straight':
			# self.console('## roishape: straight bounding box','green')
			self.shape_d = 'RECTANGLE' # dataviewer shape
			# for each contour
			for ind, itm in enumerate(contours):
				cnt = contours[ind]
				# get bounds
				_x,_y,_w,_h = cv2.boundingRect(cnt)
				# convert all coordinates floating point values to int
				roi_bounds = np.int0(cv2.boxPoints(cv2.minAreaRect(cnt)))
				#draw contours
				roi_contours = cv2.rectangle(img=image, pt1=(_x,_y), pt2=(_x+_w,_y+_h), color=color, thickness=cv2.FILLED)
			# create bounds
			_bounds = roi_bounds
			# create contours
			_contours = sum([roi_contours])

		#----rotated bounding box
		elif self.shape == 'rotate':
			# self.console('## roishape: rotated bounding box','green')
			self.shape_d = 'FREEHAND' # dataviewer shape
			# for each contour
			for ind, itm in enumerate(contours):
				cnt = contours[ind]
				# get bounds
				rect = cv2.minAreaRect(cnt)
				roi_bounds = cv2.boxPoints(rect)
				# convert all coordinates floating point values to int
				roi_bounds = np.int0(roi_bounds)
				#draw contours
				roi_contours = cv2.drawContours(image=image, contours=[roi_bounds], contourIdx=0, color=color, thickness=cv2.FILLED)
			# create bounds
			_bounds = roi_bounds
			# create contours
			_contours = sum([roi_contours])

		#----circle enclosing
		elif self.shape == 'circle':
			# self.console('## roishape: bounding circle','green')
			self.shape_d = 'ELLIPSE' # dataviewer shape
			# for each contour
			for ind, itm in enumerate(contours):
				cnt = contours[ind]
				# get minimal enclosing circle
				(_x,_y),_r = cv2.minEnclosingCircle(cnt)
				# convert all coordinates floating point values to int
				roi_bounds = np.int0(cv2.boxPoints(cv2.minAreaRect(cnt)))
				#get center and radius of circle
				center = (int(_x),int(_y))
				radius = int(_r)
				#draw contours
				roi_contours = cv2.circle(img=image, center=center, radius=radius, color=color, thickness=cv2.FILLED)
			# create bounds
			_bounds = roi_bounds
			# create contours
			_contours = sum([roi_contours])

		#----Contour Approximation
		elif self.shape == 'polygon':
			# self.console('## roishape: approximate polygon','green')
			self.shape_d = 'FREEHAND' # dataviewer shape
			# for each contour
			for ind, itm in enumerate(contours):
				cnt = contours[ind]
				epsilon = 0.01 * cv2.arcLength(cnt, True)
				# get approx polygons
				roi_bounds = cv2.approxPolyDP(cnt, epsilon, True)
				# draw approx polygons
				roi_contours = cv2.drawContours(image=image, contours=[roi_bounds], contourIdx=-1 ,color=color, thickness=cv2.FILLED)
			# create bounds
			_bounds = roi_bounds[:,0,:]
			# create contours
			_contours = sum([roi_contours])

		#----convex hull
		elif self.shape == 'hull':
			# self.console('## roishape: hull','green')
			self.shape_d = 'FREEHAND' # dataviewer shape
			# for each contour
			for ind, itm in enumerate(contours):
				cnt = contours[ind]
				# get convex hull
				roi_bounds = cv2.convexHull(itm)
				# draw hull
				roi_contours = cv2.drawContours(image=image,contours=[roi_bounds], contourIdx=-1, color=color, thickness=cv2.FILLED)
			# create bounds
			_bounds = roi_bounds[:,0,:]
			# create contours
			_contours = sum([roi_contours])

		#----no shape chosen
		else:
			raise Exception('Please select either straight, rotate, circle, polygon, box, or hull shape.')

		#self.console('test4.6', 'red')

		return _bounds, _contours

	def create_rois(self, imagename, metadata, roiname, roinumber, _bounds, _contours):
		"""[summary]

		Parameters
		----------
		imagename : [type]
			[description]
		metadata : [type]
			[description]
		roiname : [type]
			[description]
		roinumber : [type]
			[description]
		roilabel : [type]
			[description]
		_bounds : [type]
			[description]
		_contours : [type]
			[description]

		Returns
		-------
		[type]
			[description]
		[type]
			[description]

		Raises
		------
		Exception
			[description]
		"""
		#----store bounds as df
		_bounds = pd.DataFrame(_bounds)

		# transpose bounds (x0, y0, x1, y1)
		_x = _bounds[0].unique().tolist()
		_y = _bounds[1].unique().tolist()

		# check if bounding box has two x and y coordinate pairs
		if (((len(_x) == 1) or (len(_y) == 1)) and self.shape == 'straight'):
			raise Exception ("Error creating bounding box for image:roi %s:%s."%(imagename, roiname))

		# set as df
		bounds = pd.DataFrame(np.column_stack([_x[0],_y[0],_x[1],_y[1]]))

		# rename
		bounds.columns = ['x0','y0','x1','y1']

		# add index, image, roi, and shape
		bounds['image'] = imagename
		bounds['roi'] = roiname
		bounds['id'] = roinumber
		bounds['shape_d'] = self.shape_d

		# clean-up
		## convert to int
		bounds[['x0','y0','x1','y1']] = bounds[['x0','y0','x1','y1']].astype(int)

		#----save image:roi level image bounds and con
		#----save roi contours
		#!! get working: draw exact contours as coordinates
		if self.save['contours']:
			contours = _contours
		# combine metadata with coordinates
		##!!! get working

		#----save roi df
		# combine metadata with bounds
		if metadata is not None:
			bounds = pd.merge(bounds, metadata, on=['image','roi'], how='outer')

		# finish
		return bounds, contours

	def export_data(self, df, path, filename, uuid=None, newcolumn=None, level='image'):
		"""[summary]

		Parameters
		----------
		df : [type]
			Bounds.
		path : [type]
			[description]
		filename : [type]
			[description]
		uuid : [type], optional
			[description], by default None
		newcolumn : [type], optional
			[description], by default None
		nested : :obj:`string` {`image`,`all`}
			Nested order, either `image` or `all`. Default `image`.

		Returns
		-------
		[type]
			[description]
		"""
		#if workng with a single image
		if level == 'image':
			# if new column is not None and level == image
			if (isinstance(newcolumn, (dict,))):
				df[list(newcolumn.keys())[0]] = list(newcolumn.values())[0]

			# if uuid, create a unique column
			if isinstance(uuid, (list,)):
				df['uuid'] = df[uuid].apply(lambda x: ''.join(x), axis=1)
				uuid_column = 'uuid'
			# else simply use roiname
			else:
				uuid_column = self.roicolumn

		#else if workng with all images
		elif level == 'all':
			# if uuid, create a unique column
			if isinstance(uuid, (list,)):
				uuid_column = 'uuid'
			# else simply use roiname
			else:
				uuid_column = self.roicolumn

		# check if folder exists
		if not os.path.exists(path):
			os.makedirs(path)

		# export to excel
		if ((self.roi_format == 'raw') or (self.roi_format == 'both')):
			filepath = Path("%s/%s.xlsx"%(path, filename))
			df.to_excel("%s"%(filepath), index=False)

			# if debug
			if self.isDebug: self.console("## raw data saved @: %s"%(filepath),'green')

		# export to ias (dataviewer)
		if ((self.roi_format == 'dataviewer') or (self.roi_format == 'both')):
			filepath = Path("%s/%s.ias"%(path, filename))
			_bounds = '\n'.join(map(str, [
				"# EyeLink Interest Area Set created on %s."%(self.now()),
				"# Interest area set file using mdl.roi.ROI()",
				"# columns: RECTANGLE | IA number | x0 | y0 | x1 | y1 | label | color",
				"# columns: ELLIPSE | IA number | x0 | y0 | x1 | y1 | label | color",
				"# columns: FREEHAND | IA number | x0,y0 | x1,y1 | x2,y2 | x3,y3 | label | color",
				"# example: RECTANGLE 1 350 172 627 286 leftcheek red",
				"# example: ELLIPSE 2 350 172 627 286 leftcheek red",
				"# example: FREEHAND 3 350,172 627,172 627,286 350,286 leftcheek red",
				"# See Section 5.10.1 of Eyelink DataViewer Users Manual (3.2.1) for more information.",
				df[['shape_d','id','x0','y0','x1','y1',uuid_column]].to_csv(index=False, header=False).replace(',', '	')
			]))
			# save to ias
			with open("%s"%(filepath), "w") as file:
				file.write(_bounds)

			# if debug
			if self.isDebug: self.console("## dataviewer data saved @: %s"%(filepath),'green')

		return df

	def draw_image(self, imagename, data, source='bounds'):
		"""[summary]

		Parameters
		----------
		imagename : [type]
			[description]
		data : [type]
			[description]
		source : [type]
			[description]
		"""
		# draw image
		fig = plt.figure()
		ax = fig.add_subplot(111)
		## create blank image and draw to figure
		_img = Image.new("RGBA", (self.screensize[0], self.screensize[1]), (0, 0, 0, 0))
		ax.imshow(_img)
		## for each bound draw on figure
		if source=="bounds":
			for _bounds in data:
				## get bounds
				x0 = _bounds['x0'].item()
				y0 = _bounds['y0'].item()
				x1 = _bounds['x1'].item()
				y1 = _bounds['y1'].item()
				_width = x1 - x0
				_height = y1 - y0
				ax.add_patch(patches.Rectangle((x0, y0), _width, _height))
			## draw image
			_folder = '%s/img/roi/bounds'%(self.output_path)
		## for each contour draw on figure
		elif source=="contours":
			for _contours in data:
				contours = cv2.cvtColor(_contours, cv2.COLOR_BGR2RGB)
				plt.imshow(contours, zorder=1, interpolation='bilinear', alpha=1)
			## draw image
			_folder = '%s/img/roi/contours'%(self.output_path)
		## check folder
		if not os.path.exists(_folder):
			os.makedirs(_folder)
		## save
		filepath = Path('%s/%s.png'%(_folder, imagename))
		if self.isDebug: self.console('## image saved @: %s'%(filepath),'blue')
		plt.savefig(filepath, dpi=self.dpi)
		plt.close(fig)

	def process_image(self, psd):
		"""[summary]

		Parameters
		----------
		psd :  `psd_tools.PSDImage <https://psd-tools.readthedocs.io/en/latest/reference/psd_tools.html#psd_tools.PSDImage>`_
			Photoshop PSD/PSB file object. The file should include one layer for each region of interest.

		Returns
		-------
		[type]
			[description]
		"""
		## load image directly from PSD
		image = psd.topil()

		# scale image
		if self.scale != 1:
			_truesize = [image.size[0], image.size[1]]
			_scalesize = [int(_truesize[0] * self.scale), int(_truesize[1] * self.scale)]
			image = image.resize(_scalesize, Image.ANTIALIAS)
			if self.isDebug: 
				self.console('# export image','blue')
				self.console('size: %s, scaled: %s'%(_truesize, _scalesize),'green')

		# center and resize image to template screen
		background = Image.new("RGBA", (self.screensize[0], self.screensize[1]), (0, 0, 0, 0))
		offset = ((self.screensize[0] - image.size[0])//2,(self.screensize[1] - image.size[1])//2)
		background.paste(image, offset)

		return background

	def run(self, directory, core=0, queue=None):
		"""[summary]

		Parameters
		----------
		directory : :obj:`list`
			[description]
		core : :obj:`int`
			(if isMultiprocessing) Core used for this function. Default `0`.
		queue : :obj:`queue.Queue`
			Constructor for a multiprocessing 'first-in, first-out' queue. Note: Queues are thread and process safe.

		Returns
		-------
		[type]
			[description]
		"""

		#----prepare lists for all images
		l_bounds_all = []
		l_contours_all = []
		l_error = []

		#!!!----for each image
		self.console('starting()','purple')
		if self.isDebug: self.console('for each image','purple')
		for file in directory:
			# console
			if self.isDebug and self.isMultiprocessing: self.console('core: %s'%(core),'orange')

			# read image
			psd = psd_tools.PSDImage.open(file)
			imagename = os.path.splitext(os.path.basename(file))[0]
			if self.isDebug: self.console('\n# file: %s'%(imagename),'blue')

			# clear lists
			l_bounds = [] #list of bounds
			l_contours = [] #list of contours

			#!!!----for each image, save image file
			if self.save['raw']:
				#self.console('test')
				# process imaage
				image = self.process_image(psd=psd)

				# check folder
				_folder = '%s/img/'%(self.output_path)
				if not os.path.exists(_folder):
					os.makedirs(_folder)

				# figure
				fig = plt.figure()
				plt.imshow(np.array(image), zorder=1, interpolation='bilinear', alpha=1)
				plt.savefig('%s/%s.png'%(_folder, imagename), dpi=self.dpi)
				plt.close(fig)
				#self.console('test1', 'red')

			#!!!----for each region of interest
			## counter
			roinumber = 1
			## create matplotlib plot()
			## do for each roi
			for layer in psd:
				# skip if layer is main image
				if Path(layer.name).stem == imagename:
					continue
				else:
					# process metadata
					#self.console('test2', 'red')
					metadata, roiname, roilabel = self.process_metadata(imagename, layer)

					# process image
					#self.console('test3', 'red')
					image = self.process_image(layer)

					# create contours
					#self.console('test4', 'red')
					_bounds, _contours = self.create_contours(image, imagename, roiname)

					# create rois
					bounds, contours = self.create_rois(imagename, metadata, roiname, roinumber, _bounds, _contours)

					# try:
					# 	#self.console('test5', 'red')

					# except Exception as e:
					# 	# error
					# 	print(e)
					# 	#store error
					# 	l_error.append([imagename, roiname, e]) # <image><roi><error message>
					# 	# continue to next image
					# 	break

					# store processed bounds and contours to combine across image
					l_bounds.append(bounds)
					l_contours.append(contours)

					# update counter
					roinumber = roinumber + 1

			#!!!----for each image, export data
			# draw bounds
			self.draw_image(imagename, l_bounds, 'bounds')
			# draw contours
			self.draw_image(imagename, l_contours, 'contours')

			# concatinate and store bounds for all rois
			df = pd.concat(l_bounds)
			l_bounds_all.append(df)

			# export data
			_filename = "%s_bounds"%(imagename)
			_folder = '%s/data/'%(self.output_path)
			if not os.path.exists(_folder):
				os.makedirs(_folder)
			df = self.export_data(df=df, path=_folder, filename=_filename, uuid=self.uuid, newcolumn=self.newcolumn, level='image')

			# contours
			##!!! create roi file for complex shapes (not with dataviewer)

		#!!!----finished for all images
		# store
		## if multiprocessing, store in queue
		if self.isMultiprocessing:
			queue.put(l_bounds_all)
			pass
		# if not multiprocessing, return
		else:
			return l_bounds_all, l_contours_all, l_error

	def process(self):
		"""[summary]

		Returns
		-------
		[type]
			[description]
		"""
		# prepare arguements and procedure
		df = ''

		# if multiprocessing, get total cores
		if self.isMultiprocessing:
			import multiprocessing

			#----get number of available cores
			_max = multiprocessing.cpu_count() - 1

			#---check if selected max or value above possible cores
			if (self.cores == 'max') or (self.cores >= _max):
				self.cores = _max
			else:
				self.cores = self.cores

			#----double check multiproessing
			# if requested cores is 0 or 1, run without multiprocessing
			if ((self.cores == 0) or (self.cores == 1)):
				self.isMultiprocessing = False
				self.console('not multiprocessing', 'purple')
			# split directory by number of cores
			else:
				self.isMultiprocessing = True
				l_directory = np.array_split(self.directory, self.cores)
				self.console('multiprocessing with %s cores'%(self.cores), 'purple')

		# not multiprocessing
		else:
			self.isMultiprocessing = False
			self.console('not multiprocessing', 'purple')

		#----prepare to run
		# if not multiprocessing
		if not self.isMultiprocessing:
			l_bounds_all, l_contours_all, l_error = self.run(self.directory)

			# finish
			df, error = self.finished(df=l_bounds_all)

		# else if multiprocessing
		else:
			# collect each pipe (this is used to build send and recieve portions of output)
			queue = multiprocessing.Queue()

			# prepare threads
			process = [multiprocessing.Process(target=self.run, args=(l_directory[core].tolist(), core, queue,)) for core in range(self.cores)]

            # start each thread
			for p in process:
				p.daemon = True
				p.start()

			# return pipes
            # note: see https://stackoverflow.com/a/45829852
			returns = []
			for p in process:
				returns.append(queue.get())

			# wait for each process to finish
			for p in process:
				p.join()

			#----after running
			if self.isDebug: self.console('process() finished (multiprocessing)','purple')
			df, error = self.finished(returns)

		return df, error

	def finished(self, df, errors=None):
		"""
		Process bounds for all images.

		Parameters
		----------
		df : [type]
			[description]
		errors : [type], optional
			[description], by default None
		"""
		# if multiprocessing, combine df from each thread
		if self.isMultiprocessing:
			#concat data
			df = [i[0] for i in df if len(i) != 0] #check if lists are empty (i.e. if there are more threads than directories)
			df = pd.concat(df)
		# else combine lists of df to df
		else:
			df = pd.concat(df)

		#!!!----combine all rois across images
		# export to csv or dataviewer
		_folder = '%s/'%(self.output_path)
		_filename = "bounds"
		df = self.export_data(df=df, path=_folder, filename=_filename, uuid=self.uuid, level='all')

		#!!!----error log
		if bool(errors):
			_filename = Path('%s/error.csv'%(self.output_path))
			self.console("Errors found. See log %s"%(_filename), 'red')
			error = pd.DataFrame(errors, columns=['image','roi','message'])
			error.to_csv(_filename, index=False)
		else:
			error = None

		# finished
		self.console('finished()','purple')
		return df, error

#%% test
# drawing image
# img = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
# cv2.imshow('image',img)

# convert img to np array
# np_img = np.array(img)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
| @purpose: Run statistical models for analysis.  
| @date: Created on Sat May 1 15:12:38 2019  
| @author: Semeon Risom  
| @email: semeon.risom@gmail.com  
| @url: https://semeon.io/d/R33-analysis  
"""

# available classes and functions
__all__ = ['Model']

# required external libraries
__required__ = ['rpy2','pandas']

# global
from pdb import set_trace as breakpoint
import os, re
import datetime

# rpy2
#set path
os.environ["R_HOME"] = "/Library/Frameworks/R.framework/Versions/Current/Resources/"

# local libraries
if __name__ == '__main__':
	from .. import plot
	from .. import settings

class Model():
	"""Run statistical models for analysis."""
	def __init__(self, isLibrary=False):
		"""
		Initiate the mdl.r33.Model module.

        Parameters
        ----------
        isLibrary : :obj:`bool`
            Check if required libraries are available. Default `False`.
        """
        #check libraries
		if isLibrary:
			settings.library(__required__)

		# constants
		self.console = settings.console
		self.debug = settings.debug
		self.stn = settings.stn
		self.popover = settings.popover
		self.link = settings.link

	@classmethod
	def anova(cls, config, y, f, df, csv, path, effects, is_html=True):
		"""Run analysis of variance model using rpy2, seaborn and pandas.

		Parameters
		----------
		y : :obj:`str`
			Response variable.
		f : :obj:`str`
			Formula to use for analysis.
		df : :class:`pandas.DataFrame`
			Pandas dataframe of raw data.
		f : :obj:`str`
			R-compatiable formula.
		csv : :obj:`str`
			Name of generated CSV file to run analysis in R.
		path : :obj:`str`
			The directory path to save the generated files.
		effects : :obj:`list`
			List of main effects.
		is_html : :obj:`bool`
			Whether html should be generated.

		Returns
		-------
		model :  `rpy2.robjects.methods.RS4 <https://rpy2.github.io/doc/latest/html/robjects_oop.html?#rpy2.robjects.methods.RS4>`_
			Python representation of an R instance of class 'S4'.
		df_anova : :class:`pandas.DataFrame`
			Pandas dataframe of model output.
		get_anova : :class:`str`
			R script to run model.
		html : :class:`str`
			HTML output.

		Notes
		-----
		**Resources**
			- https://rpsychologist.com/r-guide-longitudinal-lme-lmer
			- https://sites.ualberta.ca/~lkgray/uploads/7/3/6/2/7362679/slides_-anova_assumptions.pdf
			- https://rpubs.com/tmcurley/twowayanova
			- https://rstudio-pubs-static.s3.amazonaws.com/158708_78d414c091fc47bd99f6f75e3bd8f4cb.html
			- https://m-clark.github.io/docs/mixedModels/anovamixed.html
			- http://dwoll.de/rexrepos/posts/anovaMixed.html
			- https://stats.stackexchange.com/questions/247582/repeated-measures-anova-in-r-errorsubject-vs-errorsubject-day
			- https://cran.r-project.org/web/packages/afex/vignettes/afex_anova_example.html#post-hoc-contrasts-and-plotting
			- http://www.let.rug.nl/nerbonne/teach/rema-stats-meth-seminar/presentations/Wieling-MixedModels-2011.pdf

		**Definition**:
			A test that allows one to make comparisons between the means of multiple groups of data, where two independent variables are considered. 

		**Assumptions of ANOVA**
			#. Normal distribution (normality)
				- Short: Samples are drawn from a normally distributed population (Q-Q Plot, Shapiro-Wilks Test)
				- Detailed Definition: Residuals in data are normally distributed.
			#. Homogeneity of variance (homoscedasticity)
				- Short: Variances are equal (or similar).
				- Detailed: Varience for a DV is constant across the sample. (residual vs fitted plot, Scale-Location plot, Levene's test)
			#. Independent observations
				- Samples have been drawn independently of each other. No analysis needed.

		**Hypothesis Interpretation**
			- **Null**: The means of all levels of an IV groups are equal.
			- **Alternative**: The mean of at least level of an IV is different.
		"""
		from rpy2.robjects import pandas2ri, r
		pandas2ri.activate()

		#----for timestamp
		_t0 = datetime.datetime.now()
		_f = debug(message='t', source="timestamp")
		console('model.anova(%s)'%(y), 'green')

		#----metadata
		source = "anova"

		#---------check if paths exist
		for path_ in [path, path+"/csv/", path+"/img/"]:
			if not os.path.exists(path_):
				os.makedirs(path_)

		#---------save data for access by R
		df.to_csv(path + "/csv/" + csv, index=None)

		#---------get number of remaining subjects (datapoints) used
		subject_anova = df.drop_duplicates(subset="participant", keep="first").shape[0]

		#---------run r
		#code
		get_anova = '\n'.join([
			'#!/usr/bin/env Rscript3.5.1',
			'rm(list=ls());',
			'#----library',
			'# core',
			'library(tidyverse);',
			'library(broom);',
			'# analysis',
			'library(lme4); library(lmerTest);',
			'# plot',
			'library(ggplot2);',
			'# estimated marginal means',
			'library(emmeans); library(multcomp);',
			'\t' + '',
			"#----repeated measures anova model", 
			'anova_ <- function(){',
			'\t' + "#----load data",
			'\t' + 'path <- "%s"'%(path + "csv/"),
			'\t' + "df <- read.csv(file=file.path(path, '%s'), header=TRUE)"%(csv),
			'\t' + '',
			'\t' + '#----set type',
			'\t' + 'df$trialType <- factor(df$trialType)',
			'\t' + 'df$cesd_group <- factor(df$cesd_group)',
			'\t' + 'df$participant <- factor(df$participant)',
			'\t' + '',
			'\t' + '#----run model',
			'\t' + 'lmer_ <- lmerTest::lmer(%s, data=df)'%(f),
			'\t' + 'aov_ <- stats::anova(lmer_)',
			'\t' + '',
			'\t' + '#----Estimated Marginal Means (Least-Squares Means) of all factor levels',
			'\t' + 'pair_ <- emmeans::emmeans(lmer_, c(%s),'%(','.join("'{0}'".format(x) for x in effects['main'].keys())),
			'\t' + '                          type = "response", adjust = "tukey")',
			'\t' + '# Comparison of Differences Between Levels of Factor',
			'\t' + 'lsmeans_ <- summary(emmeans::as.glht(pairs(pair_)), test=adjusted("free")) %>% ',
			'\t' + '   broom::tidy()',
			'\t' + '',
			'\t' + '#----getting x,y coordinates for qqplot',
			'\t' + '# create plot',
			'\t' + 'gg <- ggplot(lmer_) + ',
			'\t' + '  ggplot2::stat_qq(aes(sample = .resid, colour = factor(trialType))) + ',
			'\t' + '  ggplot2::geom_abline(linetype = "dotted")',
			'\t' + '',
			'\t' + '# convert to tibble',
			'\t' + 'gg <- ggplot_build(gg)[["data"]][[1]] %>% ',
			'\t' + '  dplyr::select(sample, theoretical)',
			'\t' + '',
			'\t' + '#----creating residuals tibble from model',
			'\t' + '# including raw data, residuals vs fitted)',
			'\t' + '# .resid=residuals, .fitted=predicted values .estimate=estimate of fixed effect',
			'\t' + 'residuals <- broom::augment(lmer_) %>% ',
			'\t' + '   dplyr::select(%s, .resid, .fitted)'%('participant' + ', ' + y + ', ' + ', '.join(effects['main'].keys())),
			'\t' + '',
			'\t' + '#----merge residuals and qq tibble',
			'\t' + 'residuals <- as.data.frame(merge(x=residuals, y=gg, by.x=".resid", by.y="sample", all.x = TRUE))',
			'\t' + '',
			'\t' + '#----prepare data for export',
			'\t' + '# convert model output to tibble',
			'\t' + 'output <- broom::tidy(aov_)',
			'\t' + '# convert model summary to tibble',
			'\t' + 'summary <- broom::glance(lmer_)',
			'\t' + '',
			'\t' + '#----return model output, model, residuals, lsmeans, model summary',
			'\t' + 'return(list(output, lmer_, residuals, lsmeans_, summary))',
		'}'])
		# load r function
		anova_r = r(get_anova)
		# run
		df_r = anova_r()

		#----extract df and model from rpy2
		df_anova = df_r[0]
		model = df_r[1]
		residuals = df_r[2]
		lsmeans = df_r[3]
		summary = df_r[4]

		#----clean data
		#rename
		df_anova = df_anova.rename(columns={'sumsq':'SS','meansq':'MS','statistic':'f','p.value':'Pr(>|f|)','DenDF':'DF','NumDF':'N'})
		# round p-value
		df_anova['Pr(>|f|)'] = df_anova[['Pr(>|f|)']].apply(lambda x: x.dropna().round(4).astype(str))
		#rename columns
		df_anova = df_anova.rename_axis("index", axis="columns")

		#----format summary
		summary = summary.rename_axis("index", axis="columns")
		summary = summary.to_html(index=True, index_names=True).replace('<table border="1" class="dataframe">',
		'<table id="table1" class="table '+source+' table-striped table-bordered hover dt-responsive nowrap"\ cellspacing="0" width="100%">')

		#----prepare metadata
		short_ = config['metadata']['short']
		long_ = config['metadata']['long']
		def_ = config['metadata']['def']
		cite_ = config['metadata']['cite']
		url_ = config['metadata']['url']
		var_ = config['metadata']['var']
		img_ = config['metadata']['img']

		#---------title, footnote, results
		title = '<b>Table 1.</b> Repeated Measures ANOVA for %s (<i>N</i> = %s).'%(long_[y], subject_anova)
		#----description
		description = ''.join([
			"<p><b>Type III Analysis of Variance Table with Satterthwaite's method</b> \
			[%a]. "%(link(name='anova', url='https://www.rdocumentation.org/packages/stats/versions/3.5.3/topics/anova')),
			"<div class='paragraph'>",
				"<div>The assumptions for the model are:</div>",
				'<ul class="number-list">',
					'<li>%s (%s).</li>'
					%(popover(name=long_['nd'],title=long_['nd'],description=def_['nd']),
					link(name='Q-Q Plot', url='#qq')),
					'<li>%s (%s).</li>'
					%(popover(name=long_['hv'],title=long_['hv'],description=def_['hv']),
					link(name='Residual vs Fitted Plot', url='#rf')),
					'<li>%s</li>'
					%(popover(name=long_['io'],title=long_['io'],description=def_['io'])),
				'</ul>',
			'</div>'
			"<div class='paragraph'>",
				"<div>The following post-hoc analysis were run:</div>",
				'<ul class="number-list">',
					'<li>%s.</li>'
					%(link(name='Estimated Marginal Means', url='#lsmean')),
					'<li>%s.</li>'
					%(link(name='Pairwise Comparison Plot', url='#boxplot')),
					'<li>%s.</li>'
					%(link(name='Q-Q Plot', url='#qq')),
					"<li>%s.</li>"
					%(link(name='Residual vs Fitted Plot', url='#rf')),
				'</ul>',
			'</div>'
		])

		#----results
		results = [
			'<div class="subtitle">Statistical Analysis</div>',
			"<p>",
			def_['exclude'],
			'We employed a repeated measures ANOVA using the anova() function in the <i>stats</i> R package \
			(%s; %s).'%(\
			popover(name=long_['R'], title=long_['R'], description=cite_['R']),
			popover(name=long_['anova'], title=long_['anova'], description=cite_['anova']))
		]

		#list of main effects    
		#"The fixed effects was %s,"%(def_['cesd_group'])
		if effects['main'] != None:
			#create qualifier #check if single random effect or multiple random effects
			qualifier = 'was included' if len(effects['main']) == 1 else 'were included'
			#start of statement
			results.append("For our model, ")
			#for each effect
			for idx, effect in enumerate(effects['main']):
				#if single item
				if len(effects['main']) == 1:
					results.append(" %s."%(popover(name=long_[effect],title=long_[effect],description=def_[effect])))
				#multiple items
				else:
					#last item
					if idx + 1 == len(effects['main']):
						results.append("and %s."%(popover(name=long_[effect],title=long_[effect],description=def_[effect])))
					else:
						results.append("%s,"%(popover(name=long_[effect],title=long_[effect],description=def_[effect])))
			#end  
			results.append("%s as main effects."%(qualifier))

		#list of random effects
		#stimulus (within subjects; %s, %s),"%(def_['iaps'], def_['pofa']),
		if effects['random'] != None:
			#create qualifier #check if single random effect or multiple random effects
			qualifier = 'was included' if len(effects['random']) == 1 else 'were included'
			#start of statement
			results.append("Random effects for")
			#for each effect
			for idx, effect in enumerate(effects['random']):
				#if single item
				if len(effects['random']) == 1:
					results.append(" %s."%(popover(name=long_[effect],title=long_[effect],description=def_[effect])))
				#multiple items
				else:
					#last item
					if idx + 1 == len(effects['random']):
						results.append("and %s."%(popover(name=long_[effect],title=long_[effect],description=def_[effect])))
					else:
						results.append("%s,"%(popover(name=long_[effect],title=long_[effect],description=def_[effect])))
			#end
			results.append('%s in the model to account for their respective variation in their slopes and intercepts.'%(qualifier))

		#outcome variable
		results.append('%s was used as the outcome measure.</p><p>'%(popover(name=long_[y],title=long_[y],description=def_[y])))

		# results
		results.append('For our analysis of %s, our results revealed'%(short_[y]))
		for idx, effect in enumerate(effects['main']):
			#get row
			row = df_anova[df_anova['term'].str.contains(effect)]
			#get term, B, se, z, p # lmer
			name = effect
			N_ = int(row['N'].values[0])
			DF_ = int(row['DF'].values[0])
			f_ = stn(row['f'].values[0])
			p_ = stn(row['Pr(>|f|)'].values[0])
			#significance
			if float(p_) <= 0.05:
				results.append('a statistically significant effect of')
			else:
				results.append('no statistically significant effect of')
			#if single item
			if len(effects['main']) == 1:
				results.append('%s (<a><i>F</i> (%s, %s) = %s, <i>p</i> = %s</a>).'%(short_[name], N_, DF_, f_, p_))
			#multiple items
			else:
				#last item
				if idx + 1 == len(effects['main']):
					results.append('%s (<a><i>F</i> (%s, %s) = %s, <i>p</i> = %s</a>).'%(short_[name], N_, DF_, f_, p_))
				else:
					results.append('%s (<a><i>F</i> (%s, %s) = %s, <i>p</i> = %s</a>),'%(short_[name], N_, DF_, f_, p_))
		## build results            
		results = ' '.join(results)

		# combine
		footnote = summary + re.sub(r'\s+', ' ', description + results).strip()

		#----create script
		script = ['<div class="code-container" style="display: none">' + '\n',
					'<div class="button-bar">'+'\n',
						'<a href="#" class="btn code hidden" source="copy" role="button">Copy</a>',
						'<a href="#" class="btn code hidden" source="download" role="button">Download</a>',
					'</div>'+'\n',
					'<pre class="line-numbers">' + '\n',
						#'<code id="editor" class="lang-r">'+'\n', #tinymce
						'<code contenteditable="true" class="lang-r" name=%s>'%(var_[y]) +'\n', #prismjs
							'%s'%(get_anova) + '\n',
						'</code>' + '\n',
					'</pre>' + '\n',
				'</div>\n']
		script = ''.join(script)

		#----plots and tables
		#lsmeans table
		#https://cran.r-project.org/web/packages/afex/vignettes/afex_anova_example.html
		try:         
			lsmeans = lsmeans.drop(['rhs'], 1)
			lsmeans['p.value'] = lsmeans[['p.value']].apply(lambda x: x.dropna().round(4).astype(str))
			lsmeans = lsmeans.rename_axis("index", axis="columns")
			lsmeans = lsmeans.rename(columns={'lhs':'Contrasts','std.error':'SE','statistic':'t','p.value':'Pr(>|t|)'})
		except:
			pass
		#create
		html_plots = []
		title_ = '<b>Table 2.</b> Pairwise Comparisons of Estimated Marginal Means.'
		footnote_ = def_['emm']
		html_plots.append({"title":title_,"footnote":footnote_,"anchor":"lsmean",'type':'table','df':lsmeans})

		# for each main effect (if categorical)
		for idx, effect in enumerate(effects['main']):
			# if categorical variable and not trialType
			if (effects['main'][effect] == 'categorical') and (effect != 'trialType'):
				#boxplot
				file = "%s_%s_boxplot.png"%(y, effect)
				path_ = path + "/img/" + file
				title_="Pairwise Comparisons Plot."
				footnote_ = 'This boxplot provides a comparison of %s and across both %s and %s stimuli. \
				Both IAPS and POFA plots were nornalized to allow direct comparison.'\
				%(short_[effect],
				popover(name=short_['iaps'], title=long_['iaps'], description=cite_['iaps'], image=img_['iaps']),
				popover(name=short_['pofa'], title=long_['pofa'], description=cite_['pofa'], image=img_['iaps']))
				# append and draw
				html_plots.append({"title":title_,"file":"%s"%(file),"footnote":footnote_, "anchor":"boxplot", 'type':'plot'})
				plot.boxplot(config=config, df=residuals, path=path_, x=effect, y=y, cat='bias')

		#probability (QQ) plot
		file = "%s_qq.png"%(y)
		path_ = path + "/img/" + file
		title_="Q-Q Plot (<a class='cat iaps'>iaps</a>, <a class='cat pofa'>pofa</a>)."
		footnote_ = def_['qq']
		html_plots.append({"title":title_,"file":"%s"%(file),"footnote":footnote_, "anchor":"qq", 'type':'plot'})
		plot.qq_plot(config=config, y=y, residuals=residuals, path=path_)

		#residuals vs fitted plot
		file = "%s_residuals.png"%(y)
		path_ = path + "/img/" + file
		title_="Residuals vs Fitted Plot (<a class='cat iaps'>iaps</a>, <a class='cat pofa'>pofa</a>)."
		footnote_ = def_['rvf']
		html_plots.append({"title":title_,"file":"%s"%(file),"footnote":footnote_, "anchor":"rf", 'type':'plot'})
		plot.residual_plot(config=config, y=y, residuals=residuals, path=path_)

		#save model and plot
		##create html
		html = None
		if is_html:
			path_ = path + '%s.html'%(y)
			html = plot.html(config=config, df=df_anova, raw_data=df, path=path_, source=source, title=title, name=y, script=script, 
					plots=html_plots, footnote=footnote, var=var_[y], short=short_[y], long=long_[y])
		#----end
		console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
		return model, df_anova, get_anova, html

	@classmethod
	def lmer(cls, config, y, f, df, exclude, csv, path, effects, is_html=True):
		"""Run linear mixed regression model, using rpy2, seaborn and pandas.

		Parameters
		----------
		y : :obj:`str`
			Response variable.
		f : :obj:`list` of :obj:`str`
			Formula to use for analysis.
		df : :class:`pandas.DataFrame`
			Pandas dataframe of raw data.
		f : :obj:`str`
			R-compatiable formula.
		exclude : :obj:`list`
			List of participants to be excluded.
		csv : :obj:`str`
			Name of generated CSV file to run analysis in R.
		path : :obj:`str`
			The directory path to save the generated files.
		effects : :obj:`list`
			List of main effects.
		is_html : :obj:`bool`
			Whether html should be generated.

		Returns
		-------
		model :  `rpy2.robjects.methods.RS4 <https://rpy2.github.io/doc/latest/html/robjects_oop.html?#rpy2.robjects.methods.RS4>`_
			Python representation of an R instance of class 'S4'.
		df_lmer : :class:`pandas.DataFrame`
			Pandas dataframe of model output.
		get_lmer : :class:`str`
			R script to run model.
		html : :class:`str`
			HTML output.

		Notes
		-----
		**Resources**
			- https://rpsychologist.com/r-guide-longitudinal-lme-lmer
			- https://stackoverflow.com/questions/47686227/poisson-regression-in-statsmodels-and-r
			- https://tsmatz.wordpress.com/2017/08/30/glm-regression-logistic-poisson-gaussian-gamma-tutorial-with-r/
			- https://stats.stackexchange.com/questions/311556/help-interpreting-count-data-glmm-using-lme4-glmer-and-glmer-nb-negative-binom
		"""
		from rpy2.robjects import pandas2ri, r
		pandas2ri.activate()

		#----for timestamp
		_t0 = datetime.datetime.now()
		_f = debug(message='t', source="timestamp")
		console('model.lmer(%s)'%(y), 'green')

		#----metadata
		source = 'onset'

		#-----exclude participants
		df_ex = df[~df['participant'].isin(exclude)]

		#----check if paths exist
		for path_ in [path, path + "csv/", path+"/img/"]:
			if not os.path.exists(path_):
				os.makedirs(path_)

		#----save data for access by R
		df_ex.to_csv(path + "csv/" + csv, index=None)

		#----get number of remaining subjects (datapoints) used
		subject_poisson = df_ex.drop_duplicates(subset="participant", keep="first").shape[0]

		#----model
		model_ = 'lmerTest::lmer(%s, data=df)'%(f)

		#----run r
		#code
		get_lmer = '\n'.join([
			'#!/usr/bin/env Rscript3.5.1',
			'rm(list=ls());',
			'#----library',
			'# core',
			'library(tidyverse);',
			'library(broom); library(broom.mixed);',
			'# analysis',
			'library(lme4); library(lmerTest);',
			'# plot',
			'library(ggplot2);',
			'# estimated marginal means',
			'library(emmeans); library(multcomp);',
			'\t' + '',
			"#----linear mixed model", 
			'lmer_ <- function(){',
			'\t' + "#----load data",
			'\t' + 'path <- "%s"'%(path + "csv/"),
			'\t' + "df <- read.csv(file=file.path(path, '%s'), header=TRUE)"%(csv),
			'\t' + '',
			'\t' + '#----drop data',
			'\t' + '# os #samples too small',
			'\t' + 'df <- df[(!df$os=="cos"),]',
			'\t' + '',
			'\t' + '#----normalize trial to [0,1] (recommended by Jason)',
			'\t' + 'df$TrialNum <- lapply(df$TrialNum, function(x){((x - 0)/(197 - 0))})',
			'\t' + '',
			'\t' + '#----set type', 
			'\t' + '# set as factor',
			'\t' + "df$os <- factor(df$os)",
			'\t' + "df$trialType <- factor(df$trialType)",
			'\t' + 'df$participant <- factor(df$participant)',
			'\t' + '# set trial as numeric (recommended by Jason)',
			'\t' + 'df$TrialNum <- as.numeric(df$TrialNum)',
			'\t' + '',      
			'\t' + '#----run model',
			'\t' + 'model <- %s'%(model_),
			'\t' + '',
			'\t' + '#----getting x,y coordinates for qqplot',
			'\t' + '# create plot',
			'\t' + 'gg <- ggplot(model) + ',
			'\t' + '  ggplot2::stat_qq(aes(sample = .resid, colour = factor(trialType))) + ',
			'\t' + '  ggplot2::geom_abline(linetype = "dotted") + ',
			'\t' + '  theme_bw()',
			'\t' + '# convert to tibble',
			'\t' + 'gg <- ggplot_build(gg)[["data"]][[1]] %>% ',
			'\t' + '  dplyr::select(sample, theoretical)',
			'\t' + '',
			'\t' + '#----creating residuals tibbles from model',
			'\t' + '# including raw data, residuals vs fitted',
			'\t' + '# .resid=residuals, .fitted=predicted values .estimate=estimate of fixed effect',
			'\t' + 'residuals <- broom::augment(model) %>% ',
			'\t' + '   dplyr::select(%s, .resid, .fitted)'%('participant' + ', sqrt.' + y + '., ' + ', '.join(effects['fixed'].keys())),
			'\t' + '',
			'\t' + '#----merge residuals and qq tibbles',
			'\t' + 'residuals <- merge(x=residuals, y=gg, by.x=".resid", by.y="sample", all.x = TRUE)',
			'\t' + '',
			'\t' + '#----prepare data for export',
			'\t' + '# convert output to tibble',
			'\t' + 'output <- broom.mixed::tidy(model)',
			'\t' + '# convert summary to tibble',
			'\t' + 'summary <- broom.mixed::glance(model)',
			'\t' + '',
			'\t' + '#----return model output, model, residuals, model summary',
			'\t' + 'return(list(output, model, residuals, summary))',
		'}'])
		#load r function
		lmer_r = r(get_lmer)
		#run
		df_r = lmer_r()

		#----extract df, model, summary from rpy2
		df_lmer = df_r[0]
		model = df_r[1]
		residuals = df_r[2]
		summary = df_r[3]

		#----clean data
		# drop row if it contains intercepts
		drop = ['sd__(Intercept)','	sd__TrialNum','cor__(Intercept).TrialNum','sd__Observation','sd__TrialNum']
		df_lmer = df_lmer[~df_lmer['term'].isin(drop)]
		#rename
		df_lmer = df_lmer.rename(columns={'estimate':'B','std.error':'SE','statistic':'t','p.value':'Pr(>|t|)'})
		# round 
		## B
		df_lmer['B'] = df_lmer[['B']].apply(lambda x: x.dropna().round(4).astype(str))
		## SE
		df_lmer['SE'] = df_lmer[['SE']].apply(lambda x: x.dropna().round(4).astype(str))
		## t
		df_lmer['t'] = df_lmer[['t']].apply(lambda x: x.dropna().round(4).astype(str))
		## p-value
		df_lmer['Pr(>|t|)'] = df_lmer[['Pr(>|t|)']].apply(lambda x: x.dropna().round(4).astype(str))
		#rename columns
		df_lmer = df_lmer.rename_axis("index", axis="columns")
		# drop column
		df_lmer = df_lmer.drop(['group','effect'], 1)

		#----format summary
		summary = summary.rename_axis("index", axis="columns")
		summary = summary.to_html(index=True, index_names=True).replace('<table border="1" class="dataframe">',
		'<table id="table2" class="table '+source+' table-striped table-bordered hover dt-responsive nowrap"\ cellspacing="0" width="100%">')

		#---------title, footnote, results
		short_ = config['metadata']['short']
		long_ = config['metadata']['long']
		def_ = config['metadata']['def']
		cite_ = config['metadata']['cite']
		url_ = config['metadata']['url']
		var_ = config['metadata']['var']

		#----title, description, and results
		title = '<b>Table 1.</b> Linear Mixed Model Regression for %s (<i>N</i> = %s).'%(long_[y], subject_poisson)
		description = ''.join([
			"<p><b>Linear Mixed Model Fit by REML (Laplace Approximation)</b> \
			[%a]. "%(link(name='lmer', url='https://www.rdocumentation.org/packages/lme4/versions/1.1-21/topics/lmer')),
			"This table summarizes effects on onset error rate with trial number, operating system, and stimulus.</p>",
			"<div class='paragraph'>",
				"<div>The assumptions for the model are:</div>",
				'<ul class="number-list">',
					'<li>%s (%s).</li>'
					%(popover(name=long_['rnd'],title=long_['rnd'],description=def_['rnd']),
					link(name='Q-Q Plot', url='#qq')),
					'<li>%s (%s).</li>'
					%(popover(name=long_['hv'],title=long_['hv'],description=def_['hv']),
					link(name='Residual vs Fitted Plot', url='#rf')),
					'<li>%s</li>'
					%(popover(name=long_['io'],title=long_['io'],description=def_['io'])),
				'</ul>',
			'</div>'
			"<div class='paragraph'>",
				"<div>The following post-hoc analysis were run:</div>",
				'<ul class="number-list">',
					'<li>%s.</li>'
					%(link(name='Individual Trend Plot', url='#itl')),
					'<li>%s.</li>'
					%(link(name='Group Trend Plot', url='#gtl')),
					'<li>%s.</li>'
					%(link(name='Q-Q Plot', url='#qq')),
					"<li>%s.</li>"
					%(link(name='Residual vs Fitted Plot', url='#rf')),
				'</ul>',
			'</div>'
		])

		#----results
		'''For statistical analysis, models were conducted with R, version 3.3.2 ([R]), using the <i>lme4</i> package [lmer]. To \
		analyze [y] we fit linear mixed-effects models using the <i>lmer</i> function. For our model, Group (NHI vs. PWA), Ambiguity 
		(ambiguous vs. unambiguous) and Context (DO-bias vs. SC-bias) were entered as fixed effects. Random intercepts and slopes 
		by participants and items were included for all fixed effects, as it was expected that participants and items would be 
		differently affected by the experimental manipulation. '''
		results = [
			'<div class="subtitle">Statistical Analysis</div>',
			"<p>",
			def_['exclude'],
			'We employed linear mixed effects models with random intercepts and slopes using the lmer() function in the <i>lme4</i> R package \
			(%s; %s).'%(\
			popover(name=long_['R'], title=long_['R'], description=cite_['R']),
			popover(name=long_['lmer'], title=long_['lmer'], description=cite_['lmer']))
		]

		#list of fixed effects    
		#"The fixed effects was %s,"%(def_['cesd_group'])
		if effects['fixed'] != None:
			#create qualifier #check if single random effect or multiple random effects
			qualifier = 'was included' if len(effects['fixed']) == 1 else 'were included'
			#start of statement
			results.append("For our model, ")
			#for each effect
			for idx, effect in enumerate(effects['fixed']):
				#if single item
				if len(effects['fixed']) == 1:
					results.append(" %s."%(popover(name=long_[effect],title=long_[effect],description=def_[effect])))
				#multiple items
				else:
					#last item
					if idx + 1 == len(effects['fixed']):
						results.append("and %s."%(popover(name=long_[effect],title=long_[effect],description=def_[effect])))
					else:
						results.append("%s,"%(popover(name=long_[effect],title=long_[effect],description=def_[effect])))
			#end  
			results.append("%s as fixed effects."%(qualifier))

		#list of random effects
		#stimulus (within subjects; %s, %s),"%(def_['iaps'], def_['pofa']),
		if effects['random'] != None:
			#create qualifier #check if single random effect or multiple random effects
			qualifier = 'was included' if len(effects['random']) == 1 else 'were included'
			#start of statement
			results.append("Random effects for")
			#for each effect
			for idx, effect in enumerate(effects['random']):
				#if single item
				if len(effects['random']) == 1:
					results.append(" %s."%(popover(name=long_[effect],title=long_[effect],description=def_[effect])))
				#multiple items
				else:
					#last item
					if idx + 1 == len(effects['random']):
						results.append("and %s."%(popover(name=long_[effect],title=long_[effect],description=def_[effect])))
					else:
						results.append("%s,"%(popover(name=long_[effect],title=long_[effect],description=def_[effect])))
			#end
			results.append('%s in the model to account for their respective variation in their slopes and intercepts.'%(qualifier))

		#outcome variable
		results.append('%s was used as the outcome measure.</p><p>'%(popover(name=long_[y],title=long_[y],description=def_[y])))

		# results
		results.append('For our analysis of %s, our results revealed'%(short_[y]))
		for idx, effect in enumerate(effects['fixed']):
			#get row
			row = df_lmer[df_lmer['term'].str.contains(effect)]
			#get term, B, se, z, p # lmer
			name = effect
			B_ = row['B'].values[0]
			se_ = row['SE'].values[0]
			p_ = row['Pr(>|t|)'].values[0]
			#significance
			if float(p_) <= 0.05:
				results.append('a statistically significant effect of')
			else:
				results.append('no statistically significant effect of')
			#if single item
			if len(effects['fixed']) == 1:
				results.append('%s (<a><i>t</i> = %s, <i>SE</i> = %s, <i>p</i> = %s</a>).'%(short_[name], B_, se_, p_))
			#multiple items
			else:
				#last item
				if idx + 1 == len(effects['fixed']):
					results.append('%s (<a><i>t</i> = %s, <i>SE</i> = %s, <i>p</i> = %s</a>).'%(short_[name], B_, se_, p_))
				else:
					results.append('%s (<a><i>t</i> = %s, <i>SE</i> = %s, <i>p</i> = %s</a>),'%(short_[name], B_, se_, p_))

		## build results 
		results = ' '.join(results)

		# combine
		footnote = summary + re.sub(r'\s+', ' ', description + results).strip()

		#---------plots
		#build plots
		plots = {}

		#---individual trend line
		clip = 250 if y=='diff_dotloc' else 200
		plots['individual'] = {}
		plots['individual']['type'] = 'plot'
		plots['individual']['file'] = '%s_individual'%(y)
		plots['individual']['path'] = path + "/img/" + "%s.png"%(plots['individual']['file'])
		plots['individual']['title'] = "Individual Trend Plot of the Difference Between Expected and True Onset Time \
		for %s (nested by subject:trial, <i>window</i> = 5)."%(long_[y])
		# footnote
		_exc = len(config['metadata']['subjects']['exclude'])
		_pct = (round(len(config['metadata']['subjects']['exclude'])/len(config['metadata']['subjects']['eyetracking']), 4)*100)
		plots['individual']['footnote'] = "Each line represents a participants individual %s across all trials. Participants with \
		Dotloc or Stimulus Onset Error median above 3SD (<i>n</i> = %s, %.1f%%) are drawn with a semi-opaque line. \
		The graph has been clipped at <i>y</i> = %s for displaying purposes."%(short_[y], _exc, _pct, clip)
		#---group trend plot
		plots['group'] = {}
		#---binned
		plots['group']["binned"] = {"bins":33,"ptype":'diff'}
		#---unbinned
		plots['group']["unbinned"] = {"bins":None,"ptype":'diff'}
		#both
		plots['group']['type'] = 'plot'
		plots['group']['file'] = '%s_group'%(y)
		plots['group']['path'] = path + "/img/" + "%s.png"%(plots['group']['file'])
		plots['group']['title'] = "Trend Plot of the Difference Between Expected and True Onset Time \
		for %s (nested by subject:trial)."%(long_[y])
		plots['group']['footnote'] = "Data is either unbinned (c,d) or binned into %s discrete evenly-sized groups (a,b). \
		The model is still fit using the original data. No participants have been excluded for this analysis. \
		The binned graph has been clipped at <i>y</i> = 1000 for displaying purposes."%(plots['group']["binned"]['bins'])
		# run
		#rename
		df = df.rename(columns={'onset>500':'onset_greater'})
		html_plots = plot.onset_diff_plot(config=config, df=df, meta=plots, drop=exclude, y=y, clip=clip)

		#---probability (QQ) plot
		file = "%s_qq.png"%(y)
		path_ = path + "/img/" + file
		title_="Q-Q Plot (<a class='cat iaps'>iaps</a>, <a class='cat pofa'>pofa</a>)."
		footnote_ = def_['qq']
		html_plots.append({"title":title_,"file":"%s"%(file),"footnote":footnote_, "anchor":"qq", 'type':'plot'})
		plot.qq_plot(config=config, y=y, residuals=residuals, path=path_)

		#---residuals vs fitted plot
		file = "%s_residuals.png"%(y)
		path_ = path + "/img/" + file
		title_="Residuals vs Fitted Plot (<a class='cat iaps'>iaps</a>, <a class='cat pofa'>pofa</a>)."
		footnote_ = def_['rvf']
		html_plots.append({"title":title_,"file":"%s"%(file),"footnote":footnote_, "anchor":"rf", 'type':'plot'})
		plot.residual_plot(config=config, y=y, residuals=residuals, path=path_)

		#---create script
		script = ['<div class="code-container" style="display: none">'+'\n',
					'<div class="button-bar">'+'\n',    
						'<a href="#" class="btn code hidden" source="copy" role="button">Copy</a>',
						'<a href="#" class="btn code hidden" source="download" role="button">Download</a>',
					'</div>'+'\n',
					'<pre class="line-numbers">'+'\n',
						#'<code id="editor" class="lang-r">'+'\n', #tinymce
						'<code contenteditable="true" class="lang-r" name=%s>'%(var_[y]) +'\n', #prismjs
						'%s\n'%(get_lmer),
						'</code>'+'\n',
					'</pre>'+'\n',         
				'</div>'+'\n']
		script = ''.join(script)

		#----html
		html = None
		if is_html:
			html_name = '%s_error'%(y)
			path_ = path + "%s.html"%(html_name)
			html = plot.html(config=config, df=df_lmer, raw_data=df, path=path_, source=source, title=title, name=html_name,
					script=script, plots=html_plots, footnote=footnote, var=var_[y], short=short_[y], long=long_[y])
		#----end
		console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
		return model, df_lmer, get_lmer, html

	@classmethod
	def logistic(cls, config, y, f, df, me, exclude, csv, path, is_html=True):
		"""Run logistic regression model, using rpy2, seaborn and pandas.

		Parameters
		----------
		y : :obj:`str`
			Response variable.
		f : :obj:`str`
			Formula to use for analysis.
		df : :class:`pandas.DataFrame`
			Pandas dataframe of raw data.
		me : :obj:`list` of :obj:`str`
			List of main effects.
		exclude : :obj:`list`
			List of participants to be excluded.
		csv : :obj:`str`
			Name of generated CSV file to run analysis in R.
		path : :obj:`str`
			The directory path to save the generated files.
		is_html : :obj:`bool`
			Whether html should be generated.

		Returns
		-------
		model :  `rpy2.robjects.methods.RS4 <https://rpy2.github.io/doc/latest/html/robjects_oop.html?#rpy2.robjects.methods.RS4>`_
			Python representation of an R instance of class 'S4'.
		df_logit : :class:`pandas.DataFrame`
			Pandas dataframe of model output.
		get_logit : :class:`str`
			R script to run model.
		html : :class:`str`
			HTML output.

		Notes
		-----
		**Resources**
			- https://rpsychologist.com/r-guide-longitudinal-lme-lmer
			- https://stats.idre.ucla.edu/r/dae/mixed-effects-logistic-regression/
			- https://www.statisticssolutions.com/assumptions-of-logistic-regression/
		"""
		from rpy2.robjects import pandas2ri, r
		pandas2ri.activate()

		#----for timestamp
		_t0 = datetime.datetime.now()
		_f = debug(message='t', source="timestamp")
		console('model.logistic(%s)'%(y), 'green')

		#----metadata
		source = 'logit'
		#get fullname
		full = y.replace("_", " ").title().replace("Cesd", "CESD")

		#-----exclude participants
		df = df[~df['participant'].isin(exclude)]

		#---------check if paths exist
		for path_ in [path, path+"/csv/", path+"/img/"]:
			if not os.path.exists(path_):
				os.makedirs(path_)

		#---------save data for access by R
		df.to_csv(path + "/csv/" + csv, index=None)

		#-------get number of remaining subjects (datapoints) used
		subjects_logit = df.drop_duplicates(subset="participant", keep="first").shape[0]

		#-------model
		#f = 'factor(cesd_group) ~ dp_bias + gaze_bias + final_gaze_bias + (1|participant)'
		# f = 'glmer(%s, \n\
		#            family=binomial(link="logit"), data=df, nAGQ=0)'%(f)
		f = 'lmer(%s, data=df)'%(f)

		#-------logit and confidence intervals to R data.frame
		##note: confidence intervals are based on the profiled log-likelihood function
		get_logit = '\n'.join([
			'#!/usr/bin/env Rscript3.5.1',
			'rm(list=ls());',
			'#----library',
			'# core',
			'library(tidyverse);',
			'library(broom);',
			'# analysis',
			'library(lme4);',
			'# plot',
			'library(ggplot2);',
			'# estimated marginal means',
			'library(emmeans); library(multcomp);',
			'\t' + '',
			'#----generalized linear mixed model',
			'logit_ <- function(){',
			'\t' + "#----load data",
			'\t' + 'path <- "%s"'%(path + "csv/"),
			'\t' + "df <- read.csv(file=file.path(path, '%s'), header=TRUE)"%(csv),
			'\t' + '',
			'\t' + '#----normalize trial to [0,1] (recommended by Jason)',
			'\t' + 'df$TrialNum <- lapply(df$TrialNum, function(x){((x - 0)/(197 - 0))})',
			'\t' + '',
			'\t' + '#----set type', 
			'\t' + '# set as factor',
			'\t' + "df$os <- factor(df$os)",
			'\t' + "df$trialType <- factor(df$trialType)",
			'\t' + 'df$participant <- factor(df$participant)',
			'\t' + '# set trial as numeric (recommended by Jason)',
			'\t' + 'df$TrialNum <- as.numeric(df$TrialNum)',
			'\t' + '',
			'\t' + '#----run model',
			'\t' + 'model <- %s'%(f),
			'\t' + '',
			'\t' + '#----odds ratio and confidence interval',
			'\t' + 'or_ci <- exp(cbind(OR=fixef(model), confint(model, parm="beta_", method="Wald"))) %>%',
			'\t' + '  cbind(term = rownames(.), .)',
			'\t' + 'rownames(or_ci) <- 1:nrow(or_ci)',
			'\t' + '',
			'\t' + '#----getting x,y coordinates for qqplot',
			'\t' + '# create plot',
			'\t' + 'gg <- ggplot(model) + ',
			'\t' + '  ggplot2::stat_qq(aes(sample = .resid, colour = factor(trialType))) + ',
			'\t' + '  ggplot2::geom_abline(linetype = "dotted") + ',
			'\t' + '  theme_bw()',
			'\t' + '# convert to tibble',
			'\t' + 'gg <- ggplot_build(gg)[["data"]][[1]] %>% ',
			'\t' + '  dplyr::select(sample, theoretical)',
			'\t' + '',
			'\t' + '#----creating residuals tibbles from model',
			'\t' + '# including raw data, residuals vs fitted',
			'\t' + '# .resid=residuals, .fitted=predicted values .estimate=estimate of fixed effect',
			'\t' + 'residuals <- broom::augment(model) %>% ',
			'\t' + '   dplyr::select(participant, trialType, cesd_group, .resid, .fitted)',
			'\t' + '',
			'\t' + '#----merge residuals and qq tibbles',
			'\t' + 'residuals <- merge(x=residuals, y=gg, by.x=".resid", by.y="sample", all.x = TRUE)',
			'\t' + '',
			'\t' + '#----prepare data for export',
			'\t' + '# convert model to tibble',
			'\t' + 'output <- broom::tidy(model)',
			'\t' + '# convert summary to tibble',
			'\t' + 'summary <- broom::glance(model)',
			'\t' + '',
			'\t' + '#----merge output with odds ratio and confidence interval',
			'\t' + 'output <- merge(x=output, y=or_ci, by.x="term", by.y="term", all.x = TRUE, sort = FALSE)',
			'\t' + '',
			'\t' + '#----return model output, model, residuals, model summary',
			'\t' + 'return(list(output, model, residuals, summary))',
		'}'])
		#----run r code
		#load r function
		logit_r = r(get_logit)
		#run
		df_r = logit_r()

		#----extract model dataframe, model, and residuals dataframe from rpy2
		df_logit = df_r[0]
		model = df_r[1]
		residuals = df_r[2]
		summary = df_r[3]

		#----clean data
		#rename
		df_logit = df_logit.rename(columns={'estimate':'B','std.error':'SE','statistic':'z','p.value':'Pr(>|z|)','97.5 %':'97.5%','2.5 %':'2.5%'})
		# round p-value
		df_logit['Pr(>|z|)'] = df_logit[['Pr(>|z|)']].apply(lambda x: x.dropna().round(4).astype(str))
		#merge ci
		df_logit['95% CI'] = df_logit[['2.5%', '97.5%']].values.tolist()
		# drop column
		df_logit = df_logit.drop(['group','effect','2.5%', '97.5%'], 1)
		# drop row if it contains intercepts
		# drop = ['sd_(Intercept).participant','sd_TrialNum.participant','cor_(Intercept).TrialNum.participant',
		#         'sd__(Intercept)','sd__TrialNum','cor__(Intercept).TrialNum']
		# df_logit = df_logit[~df_logit['term'].isin(drop)]

		#rename columns
		df_logit = df_logit.rename_axis("index", axis="columns")

		#----format summary
		summary = summary.rename_axis("index", axis="columns")
		summary = summary.to_html(index=True, index_names=True).replace('<table border="1" class="dataframe">',
		'<table id="table2" class="table '+source+' table-striped table-bordered hover dt-responsive nowrap"\ cellspacing="0" width="100%">')

		#---------title, footnote, results
		short_ = config['metadata']['short']
		long_ = config['metadata']['long']
		ref_ = config['metadata']['ref']
		def_ = config['metadata']['def']

		#-------title, footnote, and results
		title = '<b>Table 1.</b> Generalized Linear Mixed Model Regression for %s (N = %s).'%(full, subjects_logit)
		#description
		description = ''.join([
			"<p><b>Generalized Linear Mixed Model Fit by Maximum Likelihood (Adaptive Gauss-Hermite Quadrature)</b> \
			[<a class='anchor', href='https://www.rdocumentation.org/packages/lme4/versions/1.1-21/topics/glmer'>glmer</a>]. ",
			"This table summarizes effects on CESD score with trial number, bias score, and stimulus.</p>",
			"<div class='paragraph'>",
				"<div>The assumptions for the model are:</div>",
				'<ul class="number-list">',
					'<li>%s (<a href="#qq" class="anchor">Q-Q Plot</a>).</li>'%(def_['nd']),
					"<li>%s (<a href='#rf' class='anchor'>Residual vs Fitted Plot</a>).</li>"%(def_['hv']),
					'<li>%s </li>'%(def_['io']),
				'</ul>',
			'</div>'
			"<div class='paragraph'>",
				"<div>The following post-hoc analysis were run:</div>",
				'<ul class="number-list">',
					'<li><a href="#qq" class="anchor">Q-Q Plot</a>.</li>',
					"<li><a href='#rf' class='anchor'>Residual vs Fitted Plot</a>.</li>",
				'</ul>',
			'</div>'
		])
		#results
		terms = ['bias score','stimulus','trial number']
		results = [
			"The resulting data were analysed by fitting mixed effects logistic regression models in 'R', using the glmer \
			function (%s)."%(def_['lmer']),
			"The dependent variable was a binary measure of CESD score ('Low' (<16) and 'High' (≥16))."
			"The random effects were: trial and particpant.",
			"The fixed effects were: %s score (within subjects),"%(me[0].replace("_", " ")),
			"trial (within subjects; 0-197),",
			"and stimulus (within subjects; %s, %s)."%(def_['iaps'], def_['pofa']),
			"Weights were applied to the model to correct for the validity of bias score per trial."
		]
		## main effects
		for idx, effect in enumerate(me):
			#get row
			row = df_logit[df_logit['term'].str.contains(effect)]
			#get term, B, se, z, p
			term = terms[idx]
			or_ = stn(row['OR'].values[0])
			ci_ = '%s, %s'%(stn(row['95% CI'].values[0][0]),stn(row['95% CI'].values[0][1]))
			p_ = stn(row['Pr(>|z|)'].values[0])
			#append
			#if first item
			results.append('From our results, task %s did not predict the magnitude of CESD score &beta;=%s, 95%% CI[%s], p=%s.'%(term,or_,ci_,p_))

		results = '<p>' + ' '.join(results) + '</p>'

		#combine all
		footnote = re.sub(r'\s+', ' ', summary + description + results).strip()

		#----prepare script for html
		script = ['<div class="code-container" style="display: none">'+'\n',
					'<pre class="line-numbers">'+'\n',
						'<code class="lang-r">'+'\n',
						'%s\n'%(get_logit),
						'</code>'+'\n',
					'</pre>'+'\n',
				'</div>'+'\n']
		script = ''.join(script)

		#----allows seaborn to be run if matplotlib has already been loaded
		os.environ['KMP_DUPLICATE_LIB_OK']='True'

		#----plots
		html_plots = []
		#probability (QQ) plot
		file = "%s_qq.png"%(y)
		path_ = path + "/img/" + file
		title_="Q-Q Plot for CESD Group (<a class='cat iaps'>iaps</a>, <a class='cat pofa'>pofa</a>)."
		footnote_ = def_['qq']
		html_plots.append({"title":title_,"file":"%s"%(file),"footnote":footnote_, "anchor":"qq"})
		plot.qq_plot(config=config, y=y, residuals=residuals, path=path_)

		#residuals vs fitted plot
		file = "%s_residuals.png"%(y)
		path_ = path + "/img/" + file
		title_="Residuals vs Fitted for CESD Group (<a class='cat iaps'>iaps</a>, <a class='cat pofa'>pofa</a>)."
		footnote_ = def_['rvf']
		html_plots.append({"title":title_,"file":"%s"%(file),"footnote":footnote_, "anchor":"rf"})
		plot.residual_plot(config=config, y=y, residuals=residuals, path=path_)

		#----save model and plot
		#save model and plot
		##create html
		html = None
		if is_html:
			path_ = path + '/%s.html'%(y)
			html = plot.html(config=config, df=df_logit, raw_data=df, path=path_, source=source, plots=html_plots, name=y, 
					title=title, footnote=footnote, script=script)        
		#----end
		console('%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000)), 'blue')
		return model, df_logit, get_logit, html



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
| `@purpose`: Allow mdl.eyetracking.Eyelink to initiate calibration/validation/drift correction.  
| `@date`: Created on Sat May 1 15:12:38 2019  
| `@author`: Semeon Risom  
| `@email`: semeon.risom@gmail.com  
| `@url`: https://semeon.io/d/mdl
"""
# allowed imports
__all__ = ['Calibration']

# core
from pdb import set_trace as breakpoint
import os
import string
import array
from math import sin, cos, pi
from PIL import Image
import numpy as np

# local libraries
from .. import settings
from . import pylink

class Calibration(pylink.EyeLinkCustomDisplay):
	"""Allow mdl.eyetracking.Eyelink to initiate calibration/validation/drift correction."""
	def __init__(self, w, h, tracker, window):
		"""
        Initiate the mdl.eyetracking.Calibration module.

        Parameters
        ----------
        w,h : :class:`int`
            Screen width, height.
        tracker : :class:`object`
            Eyelink tracker instance.
        window :  `psychopy.visual.Window <https://www.psychopy.org/api/visual/window.html#window>`_
            PsychoPy window instance.
        """
		# import psychopy
		from psychopy import visual, event, sound

		self.visual = visual
		self.event = event
		self.sound = sound

		#---setup display
		pylink.EyeLinkCustomDisplay.__init__(self)

		#----flags
		self.is_calibration = True

		#----window
		self.window = window
		self.bg_color = self.window.color
		self.w = w
		self.h = h
		self.pylinkMinorVer = pylink.__version__.split('.')[1] # minor version 1-Mac, 11-Win/Linux

		#----check the screen units of Psychopy, forcing the screen to use 'pix'
		self.units = self.window.units
		if self.units != 'pix': self.window.setUnits('pix')

		#----mouse
		self.window.setMouseVisible(False)
		self.mouse = event.Mouse(visible=False)
		self.last_mouse_state = -1

		#----sound
		self.path = os.path.dirname(os.path.abspath(__file__)) + "\\"
		self.__target_beep__ = self.sound.Sound(self.path + "dist\\audio\\type.wav", secs=-1, loops=0)
		self.__target_beep__done__ = self.sound.Sound(self.path + "dist\\audio\\qbeep.wav", secs=-1, loops=0)
		self.__target_beep__error__ = self.sound.Sound(self.path + "dist\\audio\\error.wav", secs=-1, loops=0)

		#----color, image
		self.pal = None
		self.imgBuffInitType = 'I'
		self.img_scaling_factor = 4
		self.imagebuffer = array.array(self.imgBuffInitType)
		self.resizeImagebuffer = array.array(self.imgBuffInitType)
		self.size = (192*4, 160*4)
		self.imagebuffer = array.array(self.imgBuffInitType)
		self.resizeImagebuffer = array.array(self.imgBuffInitType)

		#----title
		self.msgHeight = self.size[1]/20.0
		self.title = visual.TextStim(win=self.window, text='', pos=(0,-self.size[1]/2-self.msgHeight), 
									units='pix', height=self.msgHeight, bold=False, color='black', 
									colorSpace='rgb', opacity=1, alignVert='center', wrapWidth=self.w)
		self.title.fontFiles = [self.path + "dist\\utils\\Helvetica.ttf"]
		self.title.font = 'Helvetica'

		#----menu
		menu = '\n'.join(['Show/Hide camera [Enter]','Switch Camera [Left, Right]',
		'Calibration [C]','Validation [V]','Continue [O]','CR [+/-]',
		'Pupil [Up/Down]','Search limit [Alt+arrows]'])
		self.menu = visual.TextStim(win=self.window, text=menu, pos=(-(self.w *.48), 0), units='pix', 
									height=38, bold=False, color='black', colorSpace='rgb', 
									opacity=1, alignHoriz='left', alignVert='center')
		self.menu.fontFiles = [self.path + "dist\\utils\\Helvetica.ttf"]
		self.menu.font = 'Helvetica'

		#----fixation
		self.line = visual.Line(win=self.window, start=(0, 0), end=(0,0), 
							lineWidth=2.0, lineColor=[0,0,0], units='pix')
		#----set circles
		self.outside = visual.Circle(win=self.window, pos=(0, 0), radius=10, fillColor=[0,0,0], 
							lineColor=[1,1,1], units='pix')
		self.inside = visual.Circle(win=self.window, pos=(0,0), radius=3, fillColor=[-1,-1,-1], 
							lineColor=[-1,-1,-1], units='pix') 

	def setup_cal_display(self):
		"""
		Shows the 'Camera Setup' screen along with menu options.

		Notes
		-----
		This function is called to setup calibration/validation display. This will be called just before we enter into 
		the calibration or validataion or drift correction mode. Any allocation per calibration or validation drift 
		correction can be done here. Also, it is normal to clear the display in this call. 

		"""
		print('setup_cal_display')
		self.window.clearBuffer()
		if self.is_calibration:
			self.menu.autoDraw = True
		self.window.flip()

	def clear_cal_display(self):
		"""Clear the 'Camera Setup' screen along with menu options."""
		print('clear_cal_display')
		self.menu.autoDraw = False
		self.title.autoDraw = False
		self.window.clearBuffer()
		self.window.color = self.bg_color
		self.window.flip()

	def exit_cal_display(self):
		"""Exit the 'Camera Setup' screen along with menu options."""
		print('exit_cal_display')
		self.window.setUnits(self.units)
		self.clear_cal_display()
		self.menu.autoDraw = False
		self.title.autoDraw = False
		#----flag
		self.is_calibration = False
    
	def record_abort_hide(self):
		"""This function is called if aborted"""
		print('record_abort_hide')

	def erase_cal_target(self):
		"""Erase the calibration/validation target."""
		print('erase_cal_target')
		self.clear_cal_display()
		#self.window.flip()
        
	def draw_cal_target(self, x, y):
		"""Draw the calibration/validation target."""
		print('draw_cal_target')
		self.clear_cal_display()
		# convert to psychopy coordinates
		x = x - (self.w / 2)
		y = -(y - (self.h / 2))

		# set calibration target position
		self.outside.pos = (x, y)
		self.inside.pos = (x, y)

		# display
		self.outside.draw()
		self.inside.draw()
		self.window.flip()

	def play_beep(self, beepid):
		""" Play a sound during calibration/drift correction."""
		if beepid == pylink.CAL_TARG_BEEP or beepid == pylink.DC_TARG_BEEP:
			self.__target_beep__.play()
		if beepid == pylink.CAL_ERR_BEEP or beepid == pylink.DC_ERR_BEEP:
			self.__target_beep__error__.play()
		if beepid in [pylink.CAL_GOOD_BEEP, pylink.DC_GOOD_BEEP]:
			self.__target_beep__done__.play()
        
	def getColorFromIndex(self, colorindex):
		"""Return psychopy colors for elements in the camera image."""
		if colorindex   ==  pylink.CR_HAIR_COLOR:          return (1,1,1)
		elif colorindex ==  pylink.PUPIL_HAIR_COLOR:       return (1,1,1)
		elif colorindex ==  pylink.PUPIL_BOX_COLOR:        return (-1,1,-1)
		elif colorindex ==  pylink.SEARCH_LIMIT_BOX_COLOR: return (1,-1,-1)
		elif colorindex ==  pylink.MOUSE_CURSOR_COLOR:     return (1,-1,-1)
		else:                                              return (0,0,0)
        
	def draw_line(self, x1, y1, x2, y2, colorindex):
		"""Draw a line. This is used for drawing crosshairs/squares."""
		y1 = (-y1  + self.size[1]/2)* self.img_scaling_factor
		x1 = (+x1  - self.size[0]/2)* self.img_scaling_factor
		y2 = (-y2  + self.size[1]/2)* self.img_scaling_factor
		x2 = (+x2  - self.size[0]/2)* self.img_scaling_factor
		color = self.getColorFromIndex(colorindex)
		self.line.start     = (x1, y1)
		self.line.end       = (x2, y2)
		self.line.lineColor = color
		self.line.draw()

	def draw_lozenge(self, x, y, width, height, colorindex):
		"""
		Draw a lozenge to show the defined search limits (x,y) is top-left corner of the bounding box.
		"""
		width = width * self.img_scaling_factor
		height = height * self.img_scaling_factor
		y = (-y + self.size[1]/2)* self.img_scaling_factor
		x = (+x - self.size[0]/2)* self.img_scaling_factor
		color = self.getColorFromIndex(colorindex)

		if width > height:
			rad = height / 2
			if rad == 0: return #cannot draw the circle with 0 radius
			Xs1 = [rad*cos(t) + x + rad for t in np.linspace(pi/2, pi/2+pi, 72)]
			Ys1 = [rad*sin(t) + y - rad for t in np.linspace(pi/2, pi/2+pi, 72)]
			Xs2 = [rad*cos(t) + x - rad + width for t in np.linspace(pi/2+pi, pi/2+2*pi, 72)]
			Ys2 = [rad*sin(t) + y - rad for t in np.linspace(pi/2+pi, pi/2+2*pi, 72)]
		else:
			rad = width / 2
			if rad == 0: return #cannot draw sthe circle with 0 radius
			Xs1 = [rad*cos(t) + x + rad for t in np.linspace(0, pi, 72)]
			Ys1 = [rad*sin(t) + y - rad for t in np.linspace(0, pi, 72)]
			Xs2 = [rad*cos(t) + x + rad for t in np.linspace(pi, 2*pi, 72)]
			Ys2 = [rad*sin(t) + y + rad - height for t in np.linspace(pi, 2*pi, 72)]

		vertices = list(zip(Xs1+Xs2, Ys1+Ys2))
		lozenge = self.visual.ShapeStim(self.window, vertices=vertices, lineWidth=2.0, lineColor=color, closeShape=True, units='pix')
		lozenge.draw()
    
	def get_mouse_state(self):
		"""Get the current mouse position and status"""
		X, Y = self.mouse.getPos()
		mX = self.size[0]/2.0*self.img_scaling_factor + X 
		mY = self.size[1]/2.0*self.img_scaling_factor - Y
		if mX <=0: mX =  0
		if mX > self.size[0]*self.img_scaling_factor:
			mX = self.size[0]*self.img_scaling_factor
		if mY < 0: mY =  0
		if mY > self.size[1]*self.img_scaling_factor:
			mY = self.size[1]*self.img_scaling_factor
		state = self.mouse.getPressed()[0] 
		mX = mX/self.img_scaling_factor
		mY = mY/self.img_scaling_factor

		if self.pylinkMinorVer == '1':
			mX = mX *2; mY = mY*2

		return ((mX, mY), state)
    
	def get_input_key(self):
		"""
		This function will be constantly pools, update the stimuli here is you need
		dynamic calibration target.
		"""
		ky=[]
		for keycode, modifier in self.event.getKeys(modifiers=True):
			k = pylink.JUNK_KEY
			if keycode   == 'f1': k = pylink.F1_KEY
			elif keycode == 'f2': k = pylink.F2_KEY
			elif keycode == 'f3': k = pylink.F3_KEY
			elif keycode == 'f4': k = pylink.F4_KEY
			elif keycode == 'f5': k = pylink.F5_KEY
			elif keycode == 'f6': k = pylink.F6_KEY
			elif keycode == 'f7': k = pylink.F7_KEY
			elif keycode == 'f8': k = pylink.F8_KEY
			elif keycode == 'f9': k = pylink.F9_KEY
			elif keycode == 'f10': k = pylink.F10_KEY
			elif keycode == 'pageup': k = pylink.PAGE_UP
			elif keycode == 'pagedown': k = pylink.PAGE_DOWN
			elif keycode == 'up': k = pylink.CURS_UP
			elif keycode == 'down': k = pylink.CURS_DOWN
			elif keycode == 'left': k = pylink.CURS_LEFT
			elif keycode == 'right': k = pylink.CURS_RIGHT
			elif keycode == 'backspace': k = ord('\b')
			elif keycode == 'return': k = pylink.ENTER_KEY
			elif keycode == 'space': k = ord(' ')
			elif keycode == 'escape': k = pylink.ESC_KEY
			elif keycode == 'tab': k = ord('\t')
			elif keycode in string.ascii_letters: k = ord(keycode)
			elif k== pylink.JUNK_KEY: k = 0

		# plus/equal & minux signs for CR adjustment
		if keycode in ['num_add', 'equal']: k = ord('+')
		if keycode in ['num_subtract', 'minus']: k = ord('-')

		if modifier['alt']==True: mod = 256
		else: mod = 0

		ky.append(pylink.KeyInput(k, mod))

		return ky

	def exit_image_display(self):
		"""Clear the camera image."""
		print('exit_image_display')

		self.clear_cal_display()
		self.menu.autoDraw = True
		self.window.flip()

	def alert_printf(self, msg):
		"""Print error messages."""
		print("Error: %s")%(msg)

	def setup_image_display(self, width, height):
		"""Set up the camera image, for newer APIs, the size is 384 x 320 pixels."""
		print('setup_image_display')

		self.last_mouse_state = -1
		self.size = ('384', '320')
		self.menu.autoDraw = True
		self.title.autoDraw = True

	def image_title(self, text):
		print('image_title')
		"""Display or update Pupil/CR info on image screen."""
		self.title.text = text

	def draw_image_line(self, width, line, totlines, buff):
		"""Display image pixel by pixel, line by line."""
		self.size = (width, totlines)

		i = 0
		for i in range(width):
			try: self.imagebuffer.append(self.pal[buff[i]])
			except: pass

		if line == totlines:
			bufferv = self.imagebuffer.tostring()
			img = Image.frombytes("RGBX", (width, totlines), bufferv)
			imgResize = img.resize((width*self.img_scaling_factor, totlines*self.img_scaling_factor))
			imgResizeVisual = self.visual.ImageStim(self.window, image=imgResize, units='pix')
			imgResizeVisual.draw()
			self.draw_cross_hair()
			self.window.flip()
			self.imagebuffer = array.array(self.imgBuffInitType)

	def set_image_palette(self, r,g,b):
		"""Given a set of RGB colors, create a list of 24bit numbers representing the pallet.
		I.e., RGB of (1,64,127) would be saved as 82047, or the number 00000001 01000000 011111111"""

		self.imagebuffer = array.array(self.imgBuffInitType)
		self.resizeImagebuffer = array.array(self.imgBuffInitType)
		sz = len(r)
		i = 0
		self.pal = []
		while i < sz:
			rf = int(b[i])
			gf = int(g[i])
			bf = int(r[i])
			self.pal.append((rf<<16) | (gf<<8) | (bf))
			i = i + 1



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
| @purpose: Eyetracking classification Module.   
| @date: Created on Sat May 1 15:12:38 2019   
| @author: Semeon Risom   
| @email: semeon.risom@gmail.com   
| @url: https://semeon.io/d/R33-analysis   
"""

# available functions
__all__ = ['ObservationModel','gaze_observation_model','gaze_transition_model','safelog','viterbi','forward_backward',
'dataset_features','transition_estimates','reestimate_observations_baum_welch','reestimate_observations_viterbi_robust',
'segment_features','classify_segments','classify_gaze']

# required external library
__required__ = ['numpy','scipy','nslr','pandas']

# core
from pdb import set_trace as breakpoint
import numpy as np
import scipy.stats
import itertools
import pandas as pd

# local libraries
if __name__ == '__main__':
	from .. import settings

FIXATION = 1
SACCADE = 2
PSO = 3
SMOOTH_PURSUIT = 4

class ObservationModel:
    """Include this - Semeon"""
    def __init__(self, dists):
        """Include this - Semeon"""
        self.classidx = {}
        self.idxclass = []
        self.dists = []
        for i, (cls, dist) in enumerate(dists.items()):
            self.idxclass.append(cls)
            self.classidx[cls] = i
            self.dists.append(dist)
        self.idxclass = np.array(self.idxclass)

    def liks(self, d):
        """Include this - Semeon"""
        scores = []
        scores = [dist.pdf(d) for dist in self.dists]
        return np.array(scores).T

    def classify(self, d):
        """Include this - Semeon"""
        return np.argmax(self.liks(d), axis=1)

    def dist(self, cls):
        """Include this - Semeon"""
        return self.dists[self.classidx[cls]]

def gaze_observation_model():
    """Include this - Semeon"""
    # Estimated from data by doi:10.3758/s13428-016-0738-9.
    params = {
        FIXATION: [[0.6039844795867605, -0.7788440631929878], [[0.1651734722683456, 0.0], [0.0, 1.5875256060544993]]],
        SACCADE: [[2.3259276064858194, 1.1333265634427712], [[0.080879690559802, 0.0], [0.0, 2.0718979621084372]]],
        PSO: [[1.7511546389160744, -1.817487032170937], [[0.0752678429860497, 0.0], [0.0, 1.356411391040218]]],
        SMOOTH_PURSUIT: [[0.8175021916433242, 0.3047120126632254], [[0.13334607025750783, 0.0], [0.0, 2.5328705587328173]]]
    }

    dists = {
        cls: scipy.stats.multivariate_normal(m, c)
        for cls, (m, c) in params.items()
    }

    return ObservationModel(dists)


def gaze_transition_model():
    """Include this - Semeon"""
    transitions = np.ones((4, 4))
    transitions[0, 2] = 0
    transitions[2, 1] = 0
    transitions[3, 2] = 0
    transitions[3, 0] = 0.5
    transitions[0, 3] = 0.5
    for i in range(len(transitions)):
        transitions[i] /= np.sum(transitions[i])
    return transitions

GazeObservationModel = gaze_observation_model()
GazeTransitionModel = gaze_transition_model()

def safelog(x):
    """Include this - Semeon"""
    return np.log10(np.clip(x, 1e-6, None))

def viterbi(initial_probs, transition_probs, emissions):
    """Include this - Semeon"""
    n_states = len(initial_probs)
    emissions = iter(emissions)
    emission = next(emissions)
    transition_probs = safelog(transition_probs)
    probs = safelog(emission) + safelog(initial_probs)
    state_stack = []

    for emission in emissions:
        emission /= np.sum(emission)
        trans_probs = transition_probs + np.row_stack(probs)
        most_likely_states = np.argmax(trans_probs, axis=0)
        probs = safelog(emission) + trans_probs[most_likely_states, np.arange(n_states)]
        state_stack.append(most_likely_states)

    state_seq = [np.argmax(probs)]

    while state_stack:
        most_likely_states = state_stack.pop()
        state_seq.append(most_likely_states[state_seq[-1]])

    state_seq.reverse()

    return state_seq

def forward_backward(transition_probs, observations, initial_probs=None):
    """Include this - Semeon"""
    observations = np.array(list(observations))
    N = len(transition_probs)
    T = len(observations)
    if initial_probs is None:
        initial_probs = np.ones(N)
        initial_probs /= np.sum(initial_probs)

    forward_probs = np.zeros((T, N))
    backward_probs = forward_probs.copy()
    probs = initial_probs
    for i in range(T):
        probs = np.dot(probs, transition_probs)*observations[i]
        probs /= np.sum(probs)
        forward_probs[i] = probs

    probs = np.ones(N)
    probs /= np.sum(probs)
    for i in range(T-1, -1, -1):
        probs = np.dot(transition_probs, (probs*observations[i]).T)
        probs /= np.sum(probs)
        backward_probs[i] = probs

    state_probs = forward_probs*backward_probs
    state_probs /= np.sum(state_probs, axis=1).reshape(-1, 1)
    return state_probs, forward_probs, backward_probs

def dataset_features(data, **nslrargs):
    segments = ((nslr.fit_gaze(ts, xs, **nslrargs), outliers) for ts, xs, outliers in data)
    features = [list(segment_features(s.segments, o)) for s, o in segments]
    return features

def transition_estimates(obs, trans, forward, backward):
    T, N = len(obs), len(trans)
    ests = np.zeros((T, N, N))
    for start, end, i in itertools.product(range(N), range(N), range(T)):
        if i == T - 1:
            b = 1/N
        else:
            b = backward[i+1, end]
        ests[i,start,end] = forward[i,start]*b*trans[start,end]
    return ests

def reestimate_observations_baum_welch(sessions,
        transition_probs=GazeTransitionModel,
        observation_model=GazeObservationModel,
        initial_probs=None,
        estimate_observation_model=True,
        estimate_transition_model=True,
        n_iterations=30,
        plot_process=False):
    """Include this - Semeon"""

    all_observations = np.vstack(sessions)

    if plot_process:
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        CLASS_COLORS = {
        1: 'b',
        2: 'r',
        3: 'y',
        4: 'g',
        5: 'm',
        6: 'c',
        22: 'orange'
        }

    for iteration in range(n_iterations): # Should probably have a nicer stopping criterion
        all_state_probs = []
        all_transition_probs = []

        # Compute state and transition probabilities
        # for all segments using the forward-backward algorithm
        for features in sessions:
            liks = np.array([observation_model.liks(f) for f in features])
            probs, forward, backward = forward_backward(transition_probs, liks, initial_probs)
            all_state_probs.extend(probs)
            all_transition_probs.append(transition_estimates(liks, transition_probs, forward, backward))

        all_state_probs = np.array(all_state_probs)
        all_transition_probs = np.vstack(all_transition_probs)
        if plot_process:
            winner = np.argmax(all_state_probs, axis=1)
            for cls in np.unique(winner):
                my = winner == cls
                plt.plot(all_observations[my,0], all_observations[my,1], '.', alpha=0.1, color=CLASS_COLORS[cls+1])
        dists = {}

        # Estimate the observation model using
        # mean and covariance of the observations weighted
        # by probability of a segment belonging to a given
        # class.
        for i, cls in enumerate(observation_model.idxclass):
            w = all_state_probs[:,i]
            wsum = np.sum(w)
            w /= wsum
            mean = np.average(all_observations, weights=w, axis=0)
            cov = np.cov(all_observations, aweights=w, rowvar=False)
            if plot_process:
                plt.plot(mean[0], mean[1], 'o', color=CLASS_COLORS[cls])
            dists[cls] = scipy.stats.multivariate_normal(mean, cov)
        if plot_process:
            plt.pause(0.1)
            plt.cla()

        if estimate_transition_model:
            # Take a mean of all sessions. This may be unoptimal.
            transition_probs = np.mean(all_transition_probs, axis=0)
            transition_probs /= np.sum(transition_probs, axis=1).reshape(-1, 1)
        if estimate_observation_model:
            observation_model=ObservationModel(dists)
    return transition_probs, observation_model

def reestimate_observations_viterbi_robust(
        sessions,
        transition_probs=GazeTransitionModel,
        observation_model=GazeObservationModel,
        initial_probs=None,
        estimate_observation_model=True,
        estimate_transition_model=True,
        n_iterations=30,
        plot_process=False):
    """Include this - Semeon"""

    from sklearn.covariance import MinCovDet
    all_observations = np.vstack(sessions)

    if plot_process:
        import matplotlib.pyplot as plt
        CLASS_COLORS = {
        1: 'b',
        2: 'r',
        3: 'y',
        4: 'g',
        5: 'm',
        6: 'c',
        22: 'orange'
        }

    N = len(transition_probs)
    if initial_probs is None:
        initial_probs = np.ones(N)
        initial_probs /= np.sum(initial_probs)

    for iteration in range(n_iterations):
        all_states = []
        all_transitions = np.zeros((N, N))
        for features in sessions:
            liks = np.array([observation_model.liks(f) for f in features])
            states = viterbi(initial_probs, transition_probs, liks)
            for i in range(len(states) - 1):
                all_transitions[states[i], states[i+1]] += 1
            all_states.extend(states)
        all_states = np.array(all_states)
        if plot_process:
            for cls in np.unique(states):
                my = all_states == cls
                plt.plot(all_observations[my,0], all_observations[my,1], '.', alpha=0.1, color=CLASS_COLORS[cls+1])

        dists = {}
        for i, cls in enumerate(observation_model.idxclass):
            my = all_states == i

            # Don't reestimate if such class is not found
            if np.sum(my) < 2:
                dists[cls] = observation_model.dists[cls]
                continue

            # Use this for non-robust
            #mean = np.average(all_observations[my], axis=0)
            #cov = observation_model.dists[i].cov

            robust = MinCovDet().fit(all_observations[my])
            mean = robust.location_
            cov = robust.covariance_
            dists[cls] = scipy.stats.multivariate_normal(mean, cov)

            if plot_process:
                plt.plot(mean[0], mean[1], 'o', color=CLASS_COLORS[cls])
        if plot_process:
            plt.pause(0.1)
            plt.cla()

        if estimate_transition_model:
            new_transition_probs = all_transitions
            totals = np.sum(new_transition_probs, axis=1).reshape(-1, 1)
            new_transition_probs /= totals

            # Avoid nans in transitions. If the algorithm
            # gets zeros here, the estimate will likely be quite bad
            not_seen = totals.flatten() == 0
            new_transition_probs[not_seen,:] = transition_probs[not_seen,:]

            transition_probs = new_transition_probs

        if estimate_observation_model:
            observation_model=ObservationModel(dists)
    return transition_probs, observation_model

def segment_features(segments, outliers=None):
    """Include this - Semeon"""
    prev_direction = np.array([0.0, 0.0])
    if outliers is None:
        outliers = np.zeros(segments[-1].i[-1], dtype=bool)
    for segment in segments:
        if np.any(outliers[segment.i[0]:segment.i[1]]): continue
        duration = float(np.diff(segment.t))
        speed = np.diff(segment.x, axis=0)/duration
        velocity = float(np.linalg.norm(speed))
        direction = speed/velocity
        cosangle = float(np.dot(direction, prev_direction.T))

        # Fisher transform, avoid exact |1|
        cosangle *= (1 - 1e-6)
        cosangle = np.arctanh(cosangle)
        if cosangle != cosangle:
            cosangle = 0.0

        yield safelog(velocity), cosangle
        prev_direction = direction

def classify_segments(segments,
        observation_model=GazeObservationModel,
        transition_model=GazeTransitionModel,
        initial_probabilities=None):
    """Include this - Semeon"""
    if initial_probabilities is None:
        initial_probabilities = np.ones(len(transition_model))
        initial_probabilities /= np.sum(initial_probabilities)
    observation_likelihoods = (observation_model.liks(f) for f in segment_features(segments))

    path = viterbi(initial_probabilities, transition_model, observation_likelihoods)
    return observation_model.idxclass[path]

def classify_gaze(ts, xs, **kwargs):
    """Include this - Semeon"""
    fit_params = {k: kwargs[k]
            for k in ('structural_error', 'optimize_noise', 'split_likelihood') if k in kwargs
    }
    #breakpoint()
    #segment data
    segmentation = nslr.fit_gaze(ts, xs, **fit_params)
    
    #classify segments #1: fixation, 2: saccade, 3: smooth pursuit, 4:PSO
    segment_classes = classify_segments(segmentation.segments)
       
    #classify samples
    sample_classes = np.zeros(len(ts))
    for c, s in zip(segment_classes, segmentation.segments):
        start = s.i[0]
        end = s.i[1]
        sample_classes[start:end] = c
    
    #prepare dataframe of segmentation data across samples ##0:start, 1:end
    l_seg = [] #list of sample data
    header = ['index start','index end','start time','end time','duration','class','cxy0','cxy1'] 
    for i, seg in enumerate(segmentation.segments):
        l_seg.append([seg.i[0],seg.i[1],
                      seg.t[0],seg.t[1],
                      seg.t[1] - seg.t[0], segment_classes[i],
                      (seg.x[0][0],seg.x[0][1]),
                      (seg.x[1][0],seg.x[0][1])
                     ])
    df = pd.DataFrame(l_seg,columns=header)
    #breakpoint() #TODO!
    
    #return sample_classes, segmentation, segment_classes, df
    return df, sample_classes









































#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@`purpose`: Sample code to run SR Research Eyelink eyetracking system. Code is optimized for the Eyelink 1000 Plus (5.0),
but should be compatiable with earlier systems.
@`date`: Created on Sat May 1 15:12:38 2019
@`author`: Semeon Risom
@`email`: semeon.risom@gmail.com
@`url`: https://semeon.io/d/mdl-R33
"""

if __name__ == "__main__":
	import os
	from setuptools import find_packages, setup, __version__

	# versioning
	import versioneer
	versioneer.VCS = "git"

	# required packages
	path = '%s/%s'%(os.path.dirname(os.path.realpath(__file__)), 'requirements.txt')
	with open(path) as f:
		required = f.read().splitlines()
	
	## add to setuptools
	setuptools_kwargs = {
		'zip_safe':True,
		'install_requires': required + ["win32api; sys_platform == 'win32'"] + ["pyobjc; sys_platform == 'darwin'"]
	}

	# setup
	name = 'imhr'
	author = 'Semeon Risom'
	packages = find_packages()
	platforms = 'Darwin, Windows'
	python_requires = '~=3.6'
	author_email = 'semeon.risom@gmail.com'
	maintainer = 'Semeon Risom'
	maintainer_email = 'semeon.risom@gmail.com'
	version = versioneer.get_version()
	cmdclass = versioneer.get_cmdclass()
	url = 'http://mdl.psy.utexas.edu/a/mdl'
	description = 'mdl - Psychology Data Science Suite.'
	download_url = 'https://github.com/risoms/mdl/'
	long_description = open('README.md').read()
	long_description_content_type = 'text/markdown'
	license_ = open('LICENSE', 'r').read()
	classifiers = [
		'Intended Audience :: Science/Research',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'License :: OSI Approved :: MIT License',
		'Topic :: Scientific/Engineering',
		'Topic :: Scientific/Engineering :: Visualization',
		'Topic :: Scientific/Engineering :: Human Machine Interfaces',
		'Topic :: Multimedia :: Graphics',
		'Operating System :: MacOS',
		'Operating System :: Microsoft :: Windows'
	]
	project_urls = {
		'Documentation': 'http://mdl.psy.utexas.edu/a/mdl',
		'Source': 'https://github.com/risoms/mdl/'
	},
	namespace_packages=['mdl']

	# init
	setup(
		name=name,
		version=version,
		packages=packages,
		platforms=platforms,
		python_requires=python_requires,
		include_package_data=True,
		author=author,
		author_email=author_email,
		maintainer=maintainer,
		maintainer_email=maintainer_email,
		description=description,
		license=license_,
		cmdclass=cmdclass,
		url=url,
		download_url=download_url,
		long_description = long_description,
		long_description_content_type = long_description_content_type,
		classifiers=classifiers,
		namespace_packages=namespace_packages,
		**setuptools_kwargs
	)
